@@ -5,10 +5,6 @@
 #ifndef SPA_DEBUG_LOG_H
 #define SPA_DEBUG_LOG_H
 
-#ifdef __cplusplus
-extern "C" {
-#endif
-
 #include <stdio.h>
 #include <stdarg.h>
 
@@ -20,6 +16,18 @@ extern "C" {
 #include <spa/debug/mem.h>
 #include <spa/debug/pod.h>
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
+#ifndef SPA_API_DEBUG_LOG
+ #ifdef SPA_API_IMPL
+  #define SPA_API_DEBUG_LOG SPA_API_IMPL
+ #else
+  #define SPA_API_DEBUG_LOG static inline
+ #endif
+#endif
+
 /**
  * \addtogroup spa_debug
  * \{
@@ -36,7 +44,7 @@ struct spa_debug_log_ctx {
 };
 
 SPA_PRINTF_FUNC(2,3)
-static inline void spa_debug_log_log(struct spa_debug_context *ctx, const char *fmt, ...)
+SPA_API_DEBUG_LOG void spa_debug_log_log(struct spa_debug_context *ctx, const char *fmt, ...)
 {
 	struct spa_debug_log_ctx *c = SPA_CONTAINER_OF(ctx, struct spa_debug_log_ctx, ctx);
 	va_list args;
