@@ -7,12 +7,17 @@
 #define SPA_AUDIO_COMPRESSED_H
 
 #include <spa/param/audio/aac.h>
+#include <spa/param/audio/ac3.h>
 #include <spa/param/audio/alac.h>
 #include <spa/param/audio/amr.h>
 #include <spa/param/audio/ape.h>
+#include <spa/param/audio/dts.h>
 #include <spa/param/audio/flac.h>
 #include <spa/param/audio/mp3.h>
+#include <spa/param/audio/mpegh.h>
+#include <spa/param/audio/opus.h>
 #include <spa/param/audio/ra.h>
+#include <spa/param/audio/truehd.h>
 #include <spa/param/audio/vorbis.h>
 #include <spa/param/audio/wma.h>
 
