@@ -5,10 +5,6 @@
 #ifndef SPA_ALSA_UTILS_H
 #define SPA_ALSA_UTILS_H
 
-#ifdef __cplusplus
-extern "C" {
-#endif
-
 #include <stddef.h>
 #include <math.h>
 
@@ -35,6 +31,9 @@ extern "C" {
 
 #include "alsa.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 #define MAX_RATES	16
 
@@ -170,6 +169,7 @@ struct state {
 	uint32_t delay;
 	uint32_t read_size;
 	uint32_t max_read;
+	uint32_t duration;
 
 	uint64_t port_info_all;
 	struct spa_port_info port_info;
@@ -231,9 +231,10 @@ struct state {
 	unsigned int is_pro:1;
 	unsigned int sources_added:1;
 	unsigned int auto_link:1;
+	unsigned int dsd_lsb:1;
 	unsigned int linked:1;
 	unsigned int is_batch:1;
-	unsigned int force_rate:1;
+	unsigned int force_quantum:1;
 	unsigned int use_period_size_min_as_headroom:1;
 
 	uint64_t iec958_codecs;
@@ -247,6 +248,7 @@ struct state {
 	uint64_t underrun;
 
 	struct spa_dll dll;
+	double dll_bw_max;
 	double max_error;
 	double max_resync;
 	double err_avg, err_var, err_wdw;
