@@ -2,6 +2,8 @@
 /* SPDX-FileCopyrightText: Copyright © 2021 Wim Taymans */
 /* SPDX-License-Identifier: MIT */
 
+#include "config.h"
+
 #include <string.h>
 #include <stdio.h>
 #include <errno.h>
@@ -13,8 +15,6 @@
 #include <limits.h>
 #include <math.h>
 
-#include "config.h"
-
 #include <spa/utils/result.h>
 #include <spa/utils/string.h>
 #include <spa/utils/json.h>
@@ -67,7 +67,7 @@ struct descriptor {
 	struct spa_list link;
 	int ref;
 	struct plugin *plugin;
-	char label[256];
+	char *label;
 
 	const struct spa_fga_descriptor *desc;
 
@@ -116,9 +116,16 @@ struct node {
 	void *hndl[MAX_HNDL];
 
 	unsigned int n_deps;
-	unsigned int visited:1;
+	uint32_t latency_index;
+
+	float min_latency;
+	float max_latency;
+
 	unsigned int disabled:1;
 	unsigned int control_changed:1;
+
+	unsigned int n_sort_deps;
+	unsigned int sorted:1;
 };
 
 struct link {
@@ -135,6 +142,7 @@ struct graph_port {
 	const struct spa_fga_descriptor *desc;
 	void **hndl;
 	uint32_t port;
+	struct node *node;
 	unsigned next:1;
 };
 
@@ -160,6 +168,7 @@ struct volume {
 struct graph {
 	struct impl *impl;
 
+	uint32_t n_nodes;
 	struct spa_list node_list;
 	struct spa_list link_list;
 
@@ -175,9 +184,26 @@ struct graph {
 	uint32_t n_control;
 	struct port **control_port;
 
+	uint32_t n_input_names;
+	char **input_names;
+
+	uint32_t n_output_names;
+	char **output_names;
+
 	struct volume volume[2];
 
+	uint32_t n_inputs;
+	uint32_t n_outputs;
+	uint32_t inputs_position[SPA_AUDIO_MAX_CHANNELS];
+	uint32_t n_inputs_position;
+	uint32_t outputs_position[SPA_AUDIO_MAX_CHANNELS];
+	uint32_t n_outputs_position;
+
+	float min_latency;
+	float max_latency;
+
 	unsigned activated:1;
+	unsigned setup:1;
 };
 
 struct impl {
@@ -205,14 +231,55 @@ struct impl {
 	float *discard_data;
 };
 
+static inline void print_channels(char *buffer, size_t max_size, uint32_t n_channels, uint32_t *positions)
+{
+	uint32_t i;
+	struct spa_strbuf buf;
+
+	spa_strbuf_init(&buf, buffer, max_size);
+	spa_strbuf_append(&buf, "[");
+	for (i = 0; i < n_channels; i++) {
+		spa_strbuf_append(&buf, "%s%s", i ? "," : "",
+			spa_type_audio_channel_to_short_name(positions[i]));
+	}
+	spa_strbuf_append(&buf, "]");
+}
+
 static void emit_filter_graph_info(struct impl *impl, bool full)
 {
 	uint64_t old = full ? impl->info.change_mask : 0;
+	struct graph *graph = &impl->graph;
 
 	if (full)
 		impl->info.change_mask = impl->info_all;
 	if (impl->info.change_mask || full) {
+		char n_inputs[64], n_outputs[64], latency[64];
+		struct spa_dict_item items[6];
+		struct spa_dict dict = SPA_DICT(items, 0);
+		char in_pos[SPA_AUDIO_MAX_CHANNELS * 8];
+		char out_pos[SPA_AUDIO_MAX_CHANNELS * 8];
+
+		snprintf(n_inputs, sizeof(n_inputs), "%d", impl->graph.n_inputs);
+		snprintf(n_outputs, sizeof(n_outputs), "%d", impl->graph.n_outputs);
+
+		items[dict.n_items++] = SPA_DICT_ITEM("n_inputs", n_inputs);
+		items[dict.n_items++] = SPA_DICT_ITEM("n_outputs", n_outputs);
+		if (graph->n_inputs_position) {
+			print_channels(in_pos, sizeof(in_pos),
+					graph->n_inputs_position, graph->inputs_position);
+			items[dict.n_items++] = SPA_DICT_ITEM("inputs.audio.position", in_pos);
+		}
+		if (graph->n_outputs_position) {
+			print_channels(out_pos, sizeof(out_pos),
+					graph->n_outputs_position, graph->outputs_position);
+			items[dict.n_items++] = SPA_DICT_ITEM("outputs.audio.position", out_pos);
+		}
+		items[dict.n_items++] = SPA_DICT_ITEM("latency",
+				spa_dtoa(latency, sizeof(latency),
+					(graph->min_latency + graph->max_latency) / 2.0f));
+		impl->info.props = &dict;
 		spa_filter_graph_emit_info(&impl->hooks, &impl->info);
+		impl->info.props = NULL;
 		impl->info.change_mask = old;
 	}
 }
@@ -243,7 +310,7 @@ static int impl_process(void *object,
 	uint32_t i, j, n_hndl = graph->n_hndl;
 	struct graph_port *port;
 
-	for (i = 0, j = 0; i < impl->info.n_inputs; i++) {
+	for (i = 0, j = 0; i < graph->n_inputs; i++) {
 		while (j < graph->n_input) {
 			port = &graph->input[j++];
 			if (port->desc && in[i])
@@ -252,7 +319,7 @@ static int impl_process(void *object,
 				break;
 		}
 	}
-	for (i = 0; i < impl->info.n_outputs; i++) {
+	for (i = 0; i < graph->n_outputs; i++) {
 		if (out[i] == NULL)
 			continue;
 
@@ -284,6 +351,15 @@ static struct node *find_node(struct graph *graph, const char *name)
 	}
 	return NULL;
 }
+#if !defined(strdupa)
+# define strdupa(s)                                                                   \
+        ({                                                                            \
+                const char *__old = (s);                                              \
+                size_t __len = strlen(__old) + 1;                                     \
+                char *__new = (char *) alloca(__len);                                 \
+                (char *) memcpy(__new, __old, __len);                                 \
+        })
+#endif
 
 /* find a port by name. Valid syntax is:
  *  "<node_name>:<port_name>"
@@ -537,11 +613,12 @@ static int parse_params(struct graph *graph, const struct spa_pod *pod)
 		return 0;
 
 	while (true) {
-		const char *name;
+		const char *name, *str_val;
 		float value, *val = NULL;
 		double dbl_val;
 		bool bool_val;
 		int32_t int_val;
+		int64_t long_val;
 
 		if (spa_pod_parser_get_string(&prs, &name) < 0)
 			break;
@@ -553,9 +630,15 @@ static int parse_params(struct graph *graph, const struct spa_pod *pod)
 		} else if (spa_pod_parser_get_int(&prs, &int_val) >= 0) {
 			value = int_val;
 			val = &value;
+		} else if (spa_pod_parser_get_long(&prs, &long_val) >= 0) {
+			value = long_val;
+			val = &value;
 		} else if (spa_pod_parser_get_bool(&prs, &bool_val) >= 0) {
 			value = bool_val ? 1.0f : 0.0f;
 			val = &value;
+		} else if (spa_pod_parser_get_string(&prs, &str_val) >= 0 &&
+			spa_json_parse_float(str_val, strlen(str_val), &value) >= 0) {
+			val = &value;
 		} else {
 			struct spa_pod *pod;
 			spa_pod_parser_get_pod(&prs, &pod);
@@ -835,6 +918,7 @@ static void descriptor_unref(struct descriptor *desc)
 	if (desc->desc)
 		spa_fga_descriptor_free(desc->desc);
 	plugin_unref(desc->plugin);
+	free(desc->label);
 	free(desc->input);
 	free(desc->output);
 	free(desc->control);
@@ -878,12 +962,12 @@ static struct descriptor *descriptor_load(struct impl *impl, const char *type,
 	spa_list_init(&desc->link);
 
 	if ((d = spa_fga_plugin_make_desc(pl->plugin, label)) == NULL) {
-		spa_log_error(impl->log, "cannot find label %s", label);
+		spa_log_error(impl->log, "cannot create label %s", label);
 		res = -ENOENT;
 		goto exit;
 	}
 	desc->desc = d;
-	snprintf(desc->label, sizeof(desc->label), "%s", label);
+	desc->label = strdup(label);
 
 	n_input = n_output = n_control = n_notify = 0;
 	for (p = 0; p < d->n_ports; p++) {
@@ -959,38 +1043,36 @@ exit:
  *   ...
  * }
  */
-static int parse_config(struct node *node, struct spa_json *config)
+static char *copy_value(struct impl *impl, struct spa_json *value)
 {
-	const char *val, *s = config->cur;
-	struct impl *impl = node->graph->impl;
-	int res = 0, len;
+	const char *val, *s = value->cur;
+	int len;
 	struct spa_error_location loc;
+	char *result = NULL;
 
-	if ((len = spa_json_next(config, &val)) <= 0) {
-		res = -EINVAL;
+	if ((len = spa_json_next(value, &val)) <= 0) {
+		errno = EINVAL;
 		goto done;
 	}
 	if (spa_json_is_null(val, len))
 		goto done;
 
 	if (spa_json_is_container(val, len)) {
-		len = spa_json_container_len(config, val, len);
+		len = spa_json_container_len(value, val, len);
 		if (len == 0) {
-			res = -EINVAL;
+			errno = EINVAL;
 			goto done;
 		}
 	}
-	if ((node->config = malloc(len+1)) == NULL) {
-		res = -errno;
+	if ((result = malloc(len+1)) == NULL)
 		goto done;
-	}
 
-	spa_json_parse_stringn(val, len, node->config, len+1);
+	spa_json_parse_stringn(val, len, result, len+1);
 done:
-	if (spa_json_get_error(config, s, &loc))
+	if (spa_json_get_error(value, s, &loc))
 		spa_debug_log_error_location(impl->log, SPA_LOG_LEVEL_WARN,
 				&loc, "error: %s", loc.reason);
-	return res;
+	return result;
 }
 
 /**
@@ -1065,7 +1147,7 @@ static int parse_link(struct graph *graph, struct spa_json *json)
 	out_port = find_port(def_out_node, output, SPA_FGA_PORT_OUTPUT);
 	in_port = find_port(def_in_node, input, SPA_FGA_PORT_INPUT);
 
-	if (out_port == NULL && out_port == NULL) {
+	if (out_port == NULL && in_port == NULL) {
 		/* try control ports */
 		out_port = find_port(def_out_node, output, SPA_FGA_PORT_OUTPUT | SPA_FGA_PORT_CONTROL);
 		in_port = find_port(def_in_node, input, SPA_FGA_PORT_INPUT | SPA_FGA_PORT_CONTROL);
@@ -1220,7 +1302,7 @@ static int parse_volume(struct graph *graph, struct spa_json *json, enum spa_dir
 static int load_node(struct graph *graph, struct spa_json *json)
 {
 	struct impl *impl = graph->impl;
-	struct spa_json control, config;
+	struct spa_json control, it;
 	struct descriptor *desc;
 	struct node *node;
 	const char *val;
@@ -1228,11 +1310,11 @@ static int load_node(struct graph *graph, struct spa_json *json)
 	char type[256] = "";
 	char name[256] = "";
 	char plugin[256] = "";
-	char label[256] = "";
+	spa_autofree char *label = NULL;
+	spa_autofree char *config = NULL;
 	bool have_control = false;
-	bool have_config = false;
 	uint32_t i;
-	int res, len;
+	int len;
 
 	while ((len = spa_json_object_next(json, key, sizeof(key), &val)) > 0) {
 		if (spa_streq("type", key)) {
@@ -1251,8 +1333,9 @@ static int load_node(struct graph *graph, struct spa_json *json)
 				return -EINVAL;
 			}
 		} else if (spa_streq("label", key)) {
-			if (spa_json_parse_stringn(val, len, label, sizeof(label)) <= 0) {
-				spa_log_error(impl->log, "label expects a string");
+			it = SPA_JSON_START(json, val);
+			if ((label = copy_value(impl, &it)) == NULL) {
+				spa_log_warn(impl->log, "error parsing label: %s", spa_strerror(-errno));
 				return -EINVAL;
 			}
 		} else if (spa_streq("control", key)) {
@@ -1263,8 +1346,9 @@ static int load_node(struct graph *graph, struct spa_json *json)
 			spa_json_enter(json, &control);
 			have_control = true;
 		} else if (spa_streq("config", key)) {
-			config = SPA_JSON_START(json, val);
-			have_config = true;
+			it = SPA_JSON_START(json, val);
+			if ((config = copy_value(impl, &it)) == NULL)
+				spa_log_warn(impl->log, "error parsing config: %s", spa_strerror(-errno));
 		} else {
 			spa_log_warn(impl->log, "unexpected node key '%s'", key);
 		}
@@ -1278,7 +1362,7 @@ static int load_node(struct graph *graph, struct spa_json *json)
 
 	spa_log_info(impl->log, "loading type:%s plugin:%s label:%s", type, plugin, label);
 
-	if ((desc = descriptor_load(graph->impl, type, plugin, label)) == NULL)
+	if ((desc = descriptor_load(graph->impl, type, plugin, label ? label : "")) == NULL)
 		return -errno;
 
 	node = calloc(1, sizeof(*node));
@@ -1288,6 +1372,8 @@ static int load_node(struct graph *graph, struct spa_json *json)
 	node->graph = graph;
 	node->desc = desc;
 	snprintf(node->name, sizeof(node->name), "%s", name);
+	node->latency_index = SPA_IDX_INVALID;
+	node->config = spa_steal_ptr(config);
 
 	node->input_port = calloc(desc->n_input, sizeof(struct port));
 	node->output_port = calloc(desc->n_output, sizeof(struct port));
@@ -1329,15 +1415,16 @@ static int load_node(struct graph *graph, struct spa_json *json)
 		port->idx = i;
 		port->external = SPA_ID_INVALID;
 		port->p = desc->notify[i];
+		if (desc->desc->ports[port->p].hint & SPA_FGA_HINT_LATENCY)
+			node->latency_index = i;
 		spa_list_init(&port->link_list);
 	}
-	if (have_config)
-		if ((res = parse_config(node, &config)) < 0)
-			spa_log_warn(impl->log, "error parsing config: %s", spa_strerror(res));
 	if (have_control)
 		parse_control(node, &control);
 
 	spa_list_append(&graph->node_list, &node->link);
+	graph->n_nodes++;
+	graph->n_control += desc->n_control;
 
 	return 0;
 }
@@ -1422,6 +1509,38 @@ static int impl_deactivate(void *object)
 	return 0;
 }
 
+static void sort_reset(struct graph *graph)
+{
+	struct node *node;
+	spa_list_for_each(node, &graph->node_list, link) {
+		node->sorted = false;
+		node->n_sort_deps = node->n_deps;
+	}
+}
+static struct node *sort_next_node(struct graph *graph)
+{
+	struct node *node;
+	spa_list_for_each(node, &graph->node_list, link) {
+		if (node->n_sort_deps == 0 && !node->sorted) {
+			uint32_t i;
+			struct link *link;
+			node->sorted = true;
+			for (i = 0; i < node->desc->n_output; i++) {
+				spa_list_for_each(link, &node->output_port[i].link_list, output_link)
+					link->input->node->n_sort_deps--;
+			}
+			for (i = 0; i < node->desc->n_notify; i++) {
+				spa_list_for_each(link, &node->notify_port[i].link_list, output_link)
+					link->input->node->n_sort_deps--;
+			}
+			return node;
+		}
+	}
+	return NULL;
+}
+
+static int setup_graph(struct graph *graph);
+
 static int impl_activate(void *object, const struct spa_dict *props)
 {
 	struct impl *impl = object;
@@ -1432,10 +1551,10 @@ static int impl_activate(void *object, const struct spa_dict *props)
 	struct descriptor *desc;
 	const struct spa_fga_descriptor *d;
 	const struct spa_fga_plugin *p;
-	uint32_t i, j, max_samples = impl->quantum_limit;
+	uint32_t i, j, max_samples = impl->quantum_limit, n_ports;
 	int res;
-	float *sd, *dd, *data;
-	const char *rate;
+	float *sd, *dd, *data, min_latency, max_latency;
+	const char *rate, *str;
 
 	if (graph->activated)
 		return 0;
@@ -1445,6 +1564,30 @@ static int impl_activate(void *object, const struct spa_dict *props)
 	rate = spa_dict_lookup(props, SPA_KEY_AUDIO_RATE);
 	impl->rate = rate ? atoi(rate) : DEFAULT_RATE;
 
+	if ((str = spa_dict_lookup(props, "filter-graph.n_inputs")) != NULL) {
+		if (spa_atou32(str, &n_ports, 0) &&
+		    n_ports != graph->n_inputs) {
+			graph->n_inputs = n_ports;
+			graph->n_outputs = 0;
+			impl->info.change_mask |= SPA_FILTER_GRAPH_CHANGE_MASK_PROPS;
+			graph->setup = false;
+		}
+	}
+	if ((str = spa_dict_lookup(props, "filter-graph.n_outputs")) != NULL) {
+		if (spa_atou32(str, &n_ports, 0) &&
+		    n_ports != graph->n_outputs) {
+			graph->n_outputs = n_ports;
+			graph->n_inputs = 0;
+			impl->info.change_mask |= SPA_FILTER_GRAPH_CHANGE_MASK_PROPS;
+			graph->setup = false;
+		}
+	}
+	if (!graph->setup) {
+		if ((res = setup_graph(graph)) < 0)
+			return res;
+		graph->setup = true;
+	}
+
 	/* first make instances */
 	spa_list_for_each(node, &graph->node_list, link) {
 		node_cleanup(node);
@@ -1483,6 +1626,8 @@ static int impl_activate(void *object, const struct spa_dict *props)
 					if ((res = port_ensure_data(link->output, i, max_samples)) < 0)
 						goto error;
 					data = link->output->audio_data[i];
+				} else if (SPA_FGA_SUPPORTS_NULL_DATA(d->ports[port->p].flags)) {
+					data = NULL;
 				} else {
 					data = sd;
 				}
@@ -1493,9 +1638,13 @@ static int impl_activate(void *object, const struct spa_dict *props)
 			for (j = 0; j < desc->n_output; j++) {
 				port = &node->output_port[j];
 				if (port->audio_data[i] == NULL) {
+					if (SPA_FGA_SUPPORTS_NULL_DATA(d->ports[port->p].flags))
+						data = NULL;
+					else
+						data = dd;
 					spa_log_info(impl->log, "connect output port %s[%d]:%s %p",
-						node->name, i, d->ports[port->p].name, dd);
-					d->connect_port(node->hndl[i], port->p, dd);
+						node->name, i, d->ports[port->p].name, data);
+					d->connect_port(node->hndl[i], port->p, data);
 				}
 			}
 			for (j = 0; j < desc->n_control; j++) {
@@ -1533,7 +1682,66 @@ static int impl_activate(void *object, const struct spa_dict *props)
 				d->control_changed(node->hndl[i]);
 		}
 	}
+	/* calculate latency */
+	sort_reset(graph);
+	while ((node = sort_next_node(graph)) != NULL) {
+		min_latency = FLT_MAX;
+		max_latency = 0.0f;
 
+		for (i = 0; i < node->desc->n_input; i++) {
+			spa_list_for_each(link, &node->input_port[i].link_list, input_link) {
+				min_latency = fminf(min_latency, link->output->node->min_latency);
+				max_latency = fmaxf(max_latency, link->output->node->max_latency);
+			}
+		}
+		min_latency = min_latency == FLT_MAX ? 0.0f : min_latency;
+
+		if (node->latency_index != SPA_IDX_INVALID) {
+			port = &node->notify_port[node->latency_index];
+			min_latency += port->control_data[0];
+			max_latency += port->control_data[0];
+
+		}
+		node->min_latency = min_latency;
+		node->max_latency = max_latency;
+		spa_log_info(impl->log, "%s latency:%f-%f", node->name, min_latency, max_latency);
+	}
+	min_latency = FLT_MAX;
+	max_latency = 0.0f;
+	for (i = 0; i < graph->n_outputs; i++) {
+		struct graph_port *port = &graph->output[i];
+		/* ports with no descriptor are ignored */
+		if (port->desc == NULL)
+			continue;
+		max_latency = fmaxf(max_latency, port->node->max_latency);
+		min_latency = fminf(min_latency, port->node->min_latency);
+	}
+	min_latency = min_latency == FLT_MAX ? 0.0f : min_latency;
+
+	spa_log_info(impl->log, "graph latency min:%f max:%f", min_latency, max_latency);
+	if (min_latency != max_latency) {
+		spa_log_warn(impl->log, "graph has unaligned latency min:%f max:%f, "
+				"consider adding delays or tweak node latency to "
+				"align the signals", min_latency, max_latency);
+		for (i = 0; i < graph->n_outputs; i++) {
+			struct graph_port *port = &graph->output[i];
+			/* port with no descriptor are ignored */
+			if (port->desc == NULL)
+				continue;
+			if (min_latency != port->node->min_latency ||
+			    max_latency != port->node->max_latency)
+				spa_log_warn(impl->log, "output port %d from %s min:%f max:%f",
+						i, port->node->name,
+						port->node->min_latency, port->node->max_latency);
+		}
+
+	}
+	if (graph->min_latency != min_latency || graph->max_latency != max_latency) {
+		graph->min_latency = min_latency;
+		graph->max_latency = max_latency;
+		impl->info.change_mask |= SPA_FILTER_GRAPH_CHANGE_MASK_PROPS;
+	}
+	emit_filter_graph_info(impl, false);
 	spa_filter_graph_emit_props_changed(&impl->hooks, SPA_DIRECTION_INPUT);
 	return 0;
 error:
@@ -1541,69 +1749,67 @@ error:
 	return res;
 }
 
-/* any default values for the controls are set in the first instance
- * of the control data. Duplicate this to the other instances now. */
-static void setup_node_controls(struct node *node)
-{
-	uint32_t i, j;
-	uint32_t n_hndl = node->n_hndl;
-	uint32_t n_ports = node->desc->n_control;
-	struct port *ports = node->control_port;
-
-	for (i = 0; i < n_ports; i++) {
-		struct port *port = &ports[i];
-		for (j = 1; j < n_hndl; j++)
-			port->control_data[j] = port->control_data[0];
-	}
-}
-
-static struct node *find_next_node(struct graph *graph)
+static void unsetup_graph(struct graph *graph)
 {
 	struct node *node;
+	uint32_t i;
+
+	free(graph->input);
+	graph->input = NULL;
+	free(graph->output);
+	graph->output = NULL;
+	free(graph->hndl);
+	graph->hndl = NULL;
+
 	spa_list_for_each(node, &graph->node_list, link) {
-		if (node->n_deps == 0 && !node->visited) {
-			node->visited = true;
-			return node;
+		struct descriptor *desc = node->desc;
+		for (i = 0; i < desc->n_input; i++) {
+			struct port *port = &node->input_port[i];
+			port->external = SPA_ID_INVALID;
+		}
+		for (i = 0; i < desc->n_output; i++) {
+			struct port *port = &node->output_port[i];
+			port->external = SPA_ID_INVALID;
 		}
 	}
-	return NULL;
 }
 
-static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_json *outputs)
+static int setup_graph(struct graph *graph)
 {
 	struct impl *impl = graph->impl;
 	struct node *node, *first, *last;
 	struct port *port;
-	struct link *link;
 	struct graph_port *gp;
 	struct graph_hndl *gh;
-	uint32_t i, j, n_nodes, n_input, n_output, n_control, n_hndl = 0;
+	uint32_t i, j, n, n_input, n_output, n_hndl = 0;
 	int res;
 	struct descriptor *desc;
 	const struct spa_fga_descriptor *d;
-	char v[256];
+	char *pname;
 	bool allow_unused;
 
+	unsetup_graph(graph);
+
 	first = spa_list_first(&graph->node_list, struct node, link);
 	last = spa_list_last(&graph->node_list, struct node, link);
 
 	/* calculate the number of inputs and outputs into the graph.
-	 * If we have a list of inputs/outputs, just count them. Otherwise
+	 * If we have a list of inputs/outputs, just use them. Otherwise
 	 * we count all input ports of the first node and all output
 	 * ports of the last node */
-	if (inputs != NULL)
-		n_input = count_array(inputs);
+	if (graph->n_input_names != 0)
+		n_input = graph->n_input_names;
 	else
 		n_input = first->desc->n_input;
 
-	if (outputs != NULL)
-		n_output = count_array(outputs);
+	if (graph->n_output_names != 0)
+		n_output = graph->n_output_names;
 	else
 		n_output = last->desc->n_output;
 
 	/* we allow unconnected ports when not explicitly given and the nodes support
 	 * NULL data */
-	allow_unused = inputs == NULL && outputs == NULL &&
+	allow_unused = graph->n_input_names == 0 && graph->n_output_names == 0 &&
 	    SPA_FLAG_IS_SET(first->desc->desc->flags, SPA_FGA_DESCRIPTOR_SUPPORTS_NULL_DATA) &&
 	    SPA_FLAG_IS_SET(last->desc->desc->flags, SPA_FGA_DESCRIPTOR_SUPPORTS_NULL_DATA);
 
@@ -1617,24 +1823,28 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 		res = -EINVAL;
 		goto error;
 	}
+	if (graph->n_inputs == 0)
+		graph->n_inputs = impl->info.n_inputs;
+	if (graph->n_inputs == 0)
+		graph->n_inputs = n_input;
 
-	if (impl->info.n_inputs == 0)
-		impl->info.n_inputs = n_input;
+	if (graph->n_outputs == 0)
+		graph->n_outputs = impl->info.n_outputs;
 
 	/* compare to the requested number of inputs and duplicate the
 	 * graph n_hndl times when needed. */
-	n_hndl = impl->info.n_inputs / n_input;
+	n_hndl = graph->n_inputs / n_input;
 
-	if (impl->info.n_outputs == 0)
-		impl->info.n_outputs = n_output * n_hndl;
+	if (graph->n_outputs == 0)
+		graph->n_outputs = n_output * n_hndl;
 
-	if (n_hndl != impl->info.n_outputs / n_output) {
+	if (n_hndl != graph->n_outputs / n_output) {
 		spa_log_error(impl->log, "invalid ports. The input stream has %1$d ports and "
 				"the filter has %2$d inputs. The output stream has %3$d ports "
 				"and the filter has %4$d outputs. input:%1$d / input:%2$d != "
 				"output:%3$d / output:%4$d. Check inputs and outputs objects.",
-				impl->info.n_inputs, n_input,
-				impl->info.n_outputs, n_output);
+				graph->n_inputs, n_input,
+				graph->n_outputs, n_output);
 		res = -EINVAL;
 		goto error;
 	}
@@ -1650,24 +1860,14 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 				"the filter has %2$d inputs. The output stream has %3$d ports "
 				"and the filter has %4$d outputs. Some filter ports will be "
 				"unconnected..",
-				impl->info.n_inputs, n_input,
-				impl->info.n_outputs, n_output);
+				graph->n_inputs, n_input,
+				graph->n_outputs, n_output);
 
-		if (impl->info.n_outputs == 0)
-			impl->info.n_outputs = n_output * n_hndl;
+		if (graph->n_outputs == 0)
+			graph->n_outputs = n_output * n_hndl;
 	}
 	spa_log_info(impl->log, "using %d instances %d %d", n_hndl, n_input, n_output);
 
-	/* now go over all nodes and create instances. */
-	n_control = 0;
-	n_nodes = 0;
-	spa_list_for_each(node, &graph->node_list, link) {
-		node->n_hndl = n_hndl;
-		desc = node->desc;
-		n_control += desc->n_control;
-		n_nodes++;
-		setup_node_controls(node);
-	}
 	graph->n_input = 0;
 	graph->input = calloc(n_input * 16 * n_hndl, sizeof(struct graph_port));
 	graph->n_output = 0;
@@ -1675,7 +1875,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 
 	/* now collect all input and output ports for all the handles. */
 	for (i = 0; i < n_hndl; i++) {
-		if (inputs == NULL) {
+		if (graph->n_input_names == 0) {
 			desc = first->desc;
 			d = desc->desc;
 			for (j = 0; j < desc->n_input; j++) {
@@ -1683,19 +1883,20 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 				spa_log_info(impl->log, "input port %s[%d]:%s",
 						first->name, i, d->ports[desc->input[j]].name);
 				gp->desc = d;
+				gp->node = first;
 				gp->hndl = &first->hndl[i];
 				gp->port = desc->input[j];
 			}
 		} else {
-			struct spa_json it = *inputs;
-			while (spa_json_get_string(&it, v, sizeof(v)) > 0) {
-				if (spa_streq(v, "null")) {
+			for (n = 0; n < graph->n_input_names; n++) {
+				pname = graph->input_names[n];
+				if (spa_streq(pname, "null")) {
 					gp = &graph->input[graph->n_input++];
 					gp->desc = NULL;
 					spa_log_info(impl->log, "ignore input port %d", graph->n_input);
-				} else if ((port = find_port(first, v, SPA_FGA_PORT_INPUT)) == NULL) {
+				} else if ((port = find_port(first, pname, SPA_FGA_PORT_INPUT)) == NULL) {
 					res = -ENOENT;
-					spa_log_error(impl->log, "input port %s not found", v);
+					spa_log_error(impl->log, "input port %s not found", pname);
 					goto error;
 				} else {
 					bool disabled = false;
@@ -1731,6 +1932,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 								peer->external = graph->n_input;
 								gp = &graph->input[graph->n_input++];
 								gp->desc = peer->node->desc->desc;
+								gp->node = peer->node;
 								gp->hndl = &peer->node->hndl[i];
 								gp->port = peer->p;
 								gp->next = true;
@@ -1747,6 +1949,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 						port->external = graph->n_input;
 						gp = &graph->input[graph->n_input++];
 						gp->desc = d;
+						gp->node = port->node;
 						gp->hndl = &port->node->hndl[i];
 						gp->port = port->p;
 						gp->next = false;
@@ -1754,7 +1957,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 				}
 			}
 		}
-		if (outputs == NULL) {
+		if (graph->n_output_names == 0) {
 			desc = last->desc;
 			d = desc->desc;
 			for (j = 0; j < desc->n_output; j++) {
@@ -1762,19 +1965,20 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 				spa_log_info(impl->log, "output port %s[%d]:%s",
 						last->name, i, d->ports[desc->output[j]].name);
 				gp->desc = d;
+				gp->node = last;
 				gp->hndl = &last->hndl[i];
 				gp->port = desc->output[j];
 			}
 		} else {
-			struct spa_json it = *outputs;
-			while (spa_json_get_string(&it, v, sizeof(v)) > 0) {
+			for (n = 0; n < graph->n_output_names; n++) {
+				pname = graph->output_names[n];
 				gp = &graph->output[graph->n_output];
-				if (spa_streq(v, "null")) {
+				if (spa_streq(pname, "null")) {
 					gp->desc = NULL;
 					spa_log_info(impl->log, "silence output port %d", graph->n_output);
-				} else if ((port = find_port(last, v, SPA_FGA_PORT_OUTPUT)) == NULL) {
+				} else if ((port = find_port(last, pname, SPA_FGA_PORT_OUTPUT)) == NULL) {
 					res = -ENOENT;
-					spa_log_error(impl->log, "output port %s not found", v);
+					spa_log_error(impl->log, "output port %s not found", pname);
 					goto error;
 				} else {
 					desc = port->node->desc;
@@ -1796,6 +2000,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 							port->node->name, i, d->ports[port->p].name);
 					port->external = graph->n_output;
 					gp->desc = d;
+					gp->node = port->node;
 					gp->hndl = &port->node->hndl[i];
 					gp->port = port->p;
 				}
@@ -1804,15 +2009,12 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 		}
 	}
 
-	/* order all nodes based on dependencies */
 	graph->n_hndl = 0;
-	graph->hndl = calloc(n_nodes * n_hndl, sizeof(struct graph_hndl));
-	graph->n_control = 0;
-	graph->control_port = calloc(n_control, sizeof(struct port *));
-	while (true) {
-		if ((node = find_next_node(graph)) == NULL)
-			break;
-
+	graph->hndl = calloc(graph->n_nodes * n_hndl, sizeof(struct graph_hndl));
+	/* order all nodes based on dependencies, first reset fields */
+	sort_reset(graph);
+	while ((node = sort_next_node(graph)) != NULL) {
+		node->n_hndl = n_hndl;
 		desc = node->desc;
 		d = desc->desc;
 
@@ -1823,19 +2025,12 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 				gh->desc = d;
 			}
 		}
-		for (i = 0; i < desc->n_output; i++) {
-			spa_list_for_each(link, &node->output_port[i].link_list, output_link)
-				link->input->node->n_deps--;
-		}
-		for (i = 0; i < desc->n_notify; i++) {
-			spa_list_for_each(link, &node->notify_port[i].link_list, output_link)
-				link->input->node->n_deps--;
-		}
-
-		/* collect all control ports on the graph */
 		for (i = 0; i < desc->n_control; i++) {
-			graph->control_port[graph->n_control] = &node->control_port[i];
-			graph->n_control++;
+			/* any default values for the controls are set in the first instance
+			 * of the control data. Duplicate this to the other instances now. */
+			struct port *port = &node->control_port[i];
+			for (j = 1; j < n_hndl; j++)
+				port->control_data[j] = port->control_data[0];
 		}
 	}
 	res = 0;
@@ -1843,6 +2038,23 @@ error:
 	return res;
 }
 
+static int setup_graph_controls(struct graph *graph)
+{
+	struct node *node;
+	uint32_t i, n_control = 0;
+
+	graph->control_port = calloc(graph->n_control, sizeof(struct port *));
+	if (graph->control_port == NULL)
+		return -errno;
+
+	spa_list_for_each(node, &graph->node_list, link) {
+		/* collect all control ports on the graph */
+		for (i = 0; i < node->desc->n_control; i++)
+			graph->control_port[n_control++] = &node->control_port[i];
+	}
+	return 0;
+}
+
 /**
  * filter.graph = {
  *     nodes = [
@@ -1900,6 +2112,26 @@ static int load_graph(struct graph *graph, const struct spa_dict *props)
 			}
 			impl->info.n_outputs = res;
 		}
+		else if (spa_streq("inputs.audio.position", key)) {
+			if (!spa_json_is_array(val, len) ||
+			    (len = spa_json_container_len(&it[0], val, len)) < 0) {
+				spa_log_error(impl->log, "%s expects an array", key);
+				return -EINVAL;
+			}
+			spa_audio_parse_position(val, len, graph->inputs_position,
+						&graph->n_inputs_position);
+			impl->info.n_inputs = graph->n_inputs_position;
+		}
+		else if (spa_streq("outputs.audio.position", key)) {
+			if (!spa_json_is_array(val, len) ||
+			    (len = spa_json_container_len(&it[0], val, len)) < 0) {
+				spa_log_error(impl->log, "%s expects an array", key);
+				return -EINVAL;
+			}
+			spa_audio_parse_position(val, len, graph->outputs_position,
+						&graph->n_outputs_position);
+			impl->info.n_outputs = graph->n_outputs_position;
+		}
 		else if (spa_streq("nodes", key)) {
 			if (!spa_json_is_array(val, len)) {
 				spa_log_error(impl->log, "%s expects an array", key);
@@ -1979,21 +2211,45 @@ static int load_graph(struct graph *graph, const struct spa_dict *props)
 				return res;
 		}
 	}
-	return setup_graph(graph, pinputs, poutputs);
+	if (pinputs != NULL) {
+		graph->n_input_names = count_array(pinputs);
+		graph->input_names = calloc(graph->n_input_names, sizeof(char *));
+		graph->n_input_names = 0;
+		while (spa_json_get_string(pinputs, key, sizeof(key)) > 0)
+			graph->input_names[graph->n_input_names++] = strdup(key);
+	}
+	if (poutputs != NULL) {
+		graph->n_output_names = count_array(poutputs);
+		graph->output_names = calloc(graph->n_output_names, sizeof(char *));
+		graph->n_output_names = 0;
+		while (spa_json_get_string(poutputs, key, sizeof(key)) > 0)
+			graph->output_names[graph->n_output_names++] = strdup(key);
+	}
+	if ((res = setup_graph_controls(graph)) < 0)
+		return res;
+	return 0;
 }
 
 static void graph_free(struct graph *graph)
 {
 	struct link *link;
 	struct node *node;
+	uint32_t i;
+
+	unsetup_graph(graph);
+
 	spa_list_consume(link, &graph->link_list, link)
 		link_free(link);
 	spa_list_consume(node, &graph->node_list, link)
 		node_free(node);
-	free(graph->input);
-	free(graph->output);
-	free(graph->hndl);
+	for (i = 0; i < graph->n_input_names; i++)
+		free(graph->input_names[i]);
+	free(graph->input_names);
+	for (i = 0; i < graph->n_output_names; i++)
+		free(graph->output_names[i]);
+	free(graph->output_names);
 	free(graph->control_port);
+	graph->control_port = NULL;
 }
 
 static const struct spa_filter_graph_methods impl_filter_graph = {
