@@ -15,10 +15,12 @@
 #include <spa/utils/names.h>
 #include <spa/utils/result.h>
 #include <spa/utils/string.h>
+#include <spa/utils/json.h>
 #include <spa/buffer/alloc.h>
 #include <spa/pod/parser.h>
 #include <spa/pod/filter.h>
 #include <spa/pod/dynamic.h>
+#include <spa/pod/simplify.h>
 #include <spa/param/param.h>
 #include <spa/param/video/format-utils.h>
 #include <spa/param/latency-utils.h>
@@ -92,21 +94,48 @@ struct impl {
 	struct spa_callbacks callbacks;
 
 	unsigned int add_listener:1;
+	unsigned int have_rate_match:1;
 	unsigned int have_format:1;
 	unsigned int recheck_format:1;
 	unsigned int started:1;
 	unsigned int ready:1;
 	unsigned int async:1;
-	unsigned int passthrough:1;
+	enum spa_param_port_config_mode mode;
 	unsigned int follower_removing:1;
 	unsigned int in_recalc;
 
 	unsigned int warned:1;
 	unsigned int driver:1;
+
+	int in_enum_sync;
 };
 
 /** \endcond */
 
+static int node_enum_params_sync(struct impl *impl, struct spa_node *node,
+		uint32_t id, uint32_t *index, const struct spa_pod *filter,
+		struct spa_pod **param, struct spa_pod_builder *builder)
+{
+	int res;
+	impl->in_enum_sync++;
+	res = spa_node_enum_params_sync(node, id, index, filter, param, builder);
+	impl->in_enum_sync--;
+	return res;
+}
+
+static int node_port_enum_params_sync(struct impl *impl, struct spa_node *node,
+		enum spa_direction direction, uint32_t port_id,
+		uint32_t id, uint32_t *index, const struct spa_pod *filter,
+		struct spa_pod **param, struct spa_pod_builder *builder)
+{
+	int res;
+	impl->in_enum_sync++;
+	res = spa_node_port_enum_params_sync(node, direction, port_id, id, index,
+			filter, param, builder);
+	impl->in_enum_sync--;
+	return res;
+}
+
 static int follower_enum_params(struct impl *this,
 				 uint32_t id,
 				 uint32_t idx,
@@ -115,19 +144,23 @@ static int follower_enum_params(struct impl *this,
 				 struct spa_pod_builder *builder)
 {
 	int res;
-	if (result->next < 0x100000 &&
-	    this->follower != this->target) {
-		if ((res = spa_node_enum_params_sync(this->target,
-				id, &result->next, filter, &result->param, builder)) == 1)
-			return res;
+	if (result->next < 0x100000) {
+		if (this->follower != this->target &&
+		    this->convert_params_flags[idx] & SPA_PARAM_INFO_READ) {
+			if ((res = node_enum_params_sync(this, this->target,
+					id, &result->next, filter, &result->param, builder)) == 1)
+				return res;
+		}
 		result->next = 0x100000;
 	}
-	if (result->next < 0x200000 && this->follower_params_flags[idx] & SPA_PARAM_INFO_READ) {
-		result->next &= 0xfffff;
-		if ((res = spa_node_enum_params_sync(this->follower,
-				id, &result->next, filter, &result->param, builder)) == 1) {
-			result->next |= 0x100000;
-			return res;
+	if (result->next < 0x200000) {
+		if (this->follower_params_flags[idx] & SPA_PARAM_INFO_READ) {
+			result->next &= 0xfffff;
+			if ((res = node_enum_params_sync(this, this->follower,
+					id, &result->next, filter, &result->param, builder)) == 1) {
+				result->next |= 0x100000;
+				return res;
+			}
 		}
 		result->next = 0x200000;
 	}
@@ -188,7 +221,7 @@ next:
 	switch (id) {
 	case SPA_PARAM_EnumPortConfig:
 	case SPA_PARAM_PortConfig:
-		if (this->passthrough) {
+		if (this->mode == SPA_PARAM_PORT_CONFIG_MODE_passthrough) {
 			switch (result.index) {
 			case 0:
 				result.param = spa_pod_builder_add_object(&b.b,
@@ -222,7 +255,7 @@ next:
 	case SPA_PARAM_Format:
 	case SPA_PARAM_Latency:
 	case SPA_PARAM_Tag:
-		res = spa_node_port_enum_params_sync(this->follower,
+		res = node_port_enum_params_sync(this, this->follower,
 				this->direction, 0,
 				id, &result.next, filter, &result.param, &b.b);
 		break;
@@ -252,7 +285,7 @@ static int link_io(struct impl *this)
 	spa_zero(this->io_rate_match);
 	this->io_rate_match.rate = 1.0;
 
-	if (this->follower == this->target) {
+	if (this->follower == this->target || !this->have_rate_match) {
 		rate_match = NULL;
 		rate_match_size = 0;
 	} else {
@@ -373,7 +406,7 @@ static int debug_params(struct impl *this, struct spa_node *node,
 	state = 0;
 	while (true) {
 		spa_pod_builder_init(&b, buffer, sizeof(buffer));
-		res = spa_node_port_enum_params_sync(node,
+		res = node_port_enum_params_sync(this, node,
 					direction, port_id,
 					id, &state,
 					NULL, &param, &b);
@@ -400,11 +433,14 @@ static int negotiate_buffers(struct impl *this)
 	struct spa_pod *param;
 	int res;
 	bool follower_alloc, conv_alloc;
-	uint32_t i, size, buffers, blocks, align, flags, stride = 0;
-	uint32_t *aligns;
+	uint32_t i, size, buffers, blocks, align, flags, stride = 0, types;
+	uint32_t *aligns, data_flags;
 	struct spa_data *datas;
 	struct spa_meta metas[1];
 	uint64_t follower_flags, conv_flags;
+	struct spa_node *alloc_node;
+	enum spa_direction alloc_direction;
+	uint32_t alloc_flags;
 
 	spa_log_debug(this->log, "%p: n_buffers:%d", this, this->n_buffers);
 
@@ -416,7 +452,7 @@ static int negotiate_buffers(struct impl *this)
 
 	state = 0;
 	param = NULL;
-	if ((res = spa_node_port_enum_params_sync(this->target,
+	if ((res = node_port_enum_params_sync(this, this->target,
 				SPA_DIRECTION_REVERSE(this->direction), 0,
 				SPA_PARAM_Buffers, &state,
 				param, &param, &b)) < 0) {
@@ -431,7 +467,7 @@ static int negotiate_buffers(struct impl *this)
 	}
 
 	state = 0;
-	if ((res = spa_node_port_enum_params_sync(this->follower,
+	if ((res = node_port_enum_params_sync(this, this->follower,
 				this->direction, 0,
 				SPA_PARAM_Buffers, &state,
 				param, &param, &b)) != 1) {
@@ -454,22 +490,23 @@ static int negotiate_buffers(struct impl *this)
 	follower_alloc = SPA_FLAG_IS_SET(follower_flags, SPA_PORT_FLAG_CAN_ALLOC_BUFFERS);
 	conv_alloc = SPA_FLAG_IS_SET(conv_flags, SPA_PORT_FLAG_CAN_ALLOC_BUFFERS);
 
-	flags = 0;
+	flags = alloc_flags = 0;
 	if (conv_alloc || follower_alloc) {
 		flags |= SPA_BUFFER_ALLOC_FLAG_NO_DATA;
-		if (conv_alloc)
-			follower_alloc = false;
+		alloc_flags = SPA_NODE_BUFFERS_FLAG_ALLOC;
 	}
 
 	align = DEFAULT_ALIGN;
+	types = SPA_ID_INVALID;
 
 	if ((res = spa_pod_parse_object(param,
 			SPA_TYPE_OBJECT_ParamBuffers, NULL,
-			SPA_PARAM_BUFFERS_buffers, SPA_POD_Int(&buffers),
-			SPA_PARAM_BUFFERS_blocks,  SPA_POD_Int(&blocks),
-			SPA_PARAM_BUFFERS_size,    SPA_POD_Int(&size),
-			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(&stride),
-			SPA_PARAM_BUFFERS_align,   SPA_POD_OPT_Int(&align))) < 0)
+			SPA_PARAM_BUFFERS_buffers,  SPA_POD_Int(&buffers),
+			SPA_PARAM_BUFFERS_blocks,   SPA_POD_Int(&blocks),
+			SPA_PARAM_BUFFERS_size,     SPA_POD_Int(&size),
+			SPA_PARAM_BUFFERS_stride,   SPA_POD_Int(&stride),
+			SPA_PARAM_BUFFERS_align,    SPA_POD_OPT_Int(&align),
+			SPA_PARAM_BUFFERS_dataType, SPA_POD_OPT_Int(&types))) < 0)
 		return res;
 
 	if (this->async)
@@ -483,9 +520,19 @@ static int negotiate_buffers(struct impl *this)
 	datas = alloca(sizeof(struct spa_data) * blocks);
 	memset(datas, 0, sizeof(struct spa_data) * blocks);
 	aligns = alloca(sizeof(uint32_t) * blocks);
+
+	data_flags = SPA_DATA_FLAG_READWRITE;
+	if (SPA_FLAG_IS_SET(follower_flags, SPA_PORT_FLAG_DYNAMIC_DATA) &&
+	    SPA_FLAG_IS_SET(conv_flags, SPA_PORT_FLAG_DYNAMIC_DATA))
+		data_flags |= SPA_DATA_FLAG_DYNAMIC;
+
+	/* if we allocate, we allocate MemPtr memory */
+	if (!SPA_FLAG_IS_SET(alloc_flags, SPA_NODE_BUFFERS_FLAG_ALLOC))
+		types = SPA_DATA_MemPtr;
+
 	for (i = 0; i < blocks; i++) {
-		datas[i].type = SPA_DATA_MemPtr;
-		datas[i].flags = SPA_DATA_FLAG_READWRITE | SPA_DATA_FLAG_DYNAMIC;
+		datas[i].type = types;
+		datas[i].flags = data_flags;
 		datas[i].maxsize = size;
 		aligns[i] = align;
 	}
@@ -498,15 +545,26 @@ static int negotiate_buffers(struct impl *this)
 		return -errno;
 	this->n_buffers = buffers;
 
-	if ((res = spa_node_port_use_buffers(this->target,
-		       SPA_DIRECTION_REVERSE(this->direction), 0,
-		       conv_alloc ? SPA_NODE_BUFFERS_FLAG_ALLOC : 0,
+	/* prefer to let the follower alloc */
+	if (follower_alloc) {
+		alloc_node = this->follower;
+		alloc_direction = this->direction;
+	} else {
+		alloc_node = this->target;
+		alloc_direction = SPA_DIRECTION_REVERSE(this->direction);
+	}
+
+	if ((res = spa_node_port_use_buffers(alloc_node,
+		       alloc_direction, 0, alloc_flags,
 		       this->buffers, this->n_buffers)) < 0)
 		return res;
 
-	if ((res = spa_node_port_use_buffers(this->follower,
-		       this->direction, 0,
-		       follower_alloc ? SPA_NODE_BUFFERS_FLAG_ALLOC : 0,
+	alloc_node = alloc_node == this->follower ? this->target : this->follower;
+	alloc_direction = SPA_DIRECTION_REVERSE(alloc_direction);
+	alloc_flags = 0;
+
+	if ((res = spa_node_port_use_buffers(alloc_node,
+		       alloc_direction, 0, alloc_flags,
 		       this->buffers, this->n_buffers)) < 0)
 		return res;
 
@@ -551,7 +609,7 @@ static int configure_format(struct impl *this, uint32_t flags, const struct spa_
 
 		/* format was changed to nearest compatible format */
 
-		if ((res = spa_node_port_enum_params_sync(this->follower,
+		if ((res = node_port_enum_params_sync(this, this->follower,
 					this->direction, 0,
 					SPA_PARAM_Format, &state,
 					NULL, &fmt, &b)) != 1)
@@ -617,7 +675,7 @@ static int recalc_latency(struct impl *this, struct spa_node *src, enum spa_dire
 
 	while (true) {
 		spa_pod_builder_init(&b, buffer, sizeof(buffer));
-		if ((res = spa_node_port_enum_params_sync(src,
+		if ((res = node_port_enum_params_sync(this, src,
 						direction, port_id, SPA_PARAM_Latency,
 						&index, NULL, &param, &b)) != 1) {
 			param = NULL;
@@ -658,7 +716,7 @@ static int recalc_tag(struct impl *this, struct spa_node *src, enum spa_directio
 	while (true) {
 		void *tag_state = NULL;
 		spa_pod_builder_reset(&b.b, &state);
-		if ((res = spa_node_port_enum_params_sync(src,
+		if ((res = node_port_enum_params_sync(this, src,
 						direction, port_id, SPA_PARAM_Tag,
 						&index, NULL, &param, &b.b)) != 1) {
 			param = NULL;
@@ -680,13 +738,14 @@ static int reconfigure_mode(struct impl *this, enum spa_param_port_config_mode m
 	int res = 0;
 	struct spa_hook l;
 	bool passthrough = mode == SPA_PARAM_PORT_CONFIG_MODE_passthrough;
+	bool old_passthrough = this->mode == SPA_PARAM_PORT_CONFIG_MODE_passthrough;
 
 	spa_log_debug(this->log, "%p: passthrough mode %d", this, passthrough);
 
 	if (!passthrough && this->convert == NULL)
 		return -ENOTSUP;
 
-	if (this->passthrough != passthrough) {
+	if (old_passthrough != passthrough) {
 		if (passthrough) {
 			/* remove converter split/merge ports */
 			configure_convert(this, SPA_PARAM_PORT_CONFIG_MODE_none);
@@ -706,21 +765,22 @@ static int reconfigure_mode(struct impl *this, enum spa_param_port_config_mode m
 	if ((res = configure_format(this, SPA_NODE_PARAM_FLAG_NEAREST, format)) < 0)
 		return res;
 
-	if (this->passthrough != passthrough) {
-		this->passthrough = passthrough;
-		if (passthrough) {
-			/* add follower ports */
-			spa_zero(l);
-			spa_node_add_listener(this->follower, &l, &follower_node_events, this);
-			spa_hook_remove(&l);
-		} else {
-			/* add converter ports */
-			configure_convert(this, mode);
-		}
-		link_io(this);
+	this->mode = mode;
+
+	if (old_passthrough != passthrough && passthrough) {
+		/* add follower ports */
+		spa_zero(l);
+		spa_node_add_listener(this->follower, &l, &follower_node_events, this);
+		spa_hook_remove(&l);
+	} else {
+		/* add converter ports */
+		configure_convert(this, mode);
 	}
+	link_io(this);
+
 	this->info.change_mask |= SPA_NODE_CHANGE_MASK_FLAGS | SPA_NODE_CHANGE_MASK_PARAMS;
-	SPA_FLAG_CLEAR(this->info.flags, SPA_NODE_FLAG_NEED_CONFIGURE);
+	SPA_FLAG_UPDATE(this->info.flags, SPA_NODE_FLAG_NEED_CONFIGURE,
+			this->mode == SPA_PARAM_PORT_CONFIG_MODE_none);
 	SPA_FLAG_UPDATE(this->info.flags, SPA_NODE_FLAG_ASYNC,
 			this->async && this->follower == this->target);
 	this->params[IDX_Props].user++;
@@ -854,6 +914,54 @@ static int impl_node_set_io(void *object, uint32_t id, void *data, size_t size)
 	return res;
 }
 
+static int update_param_peer_formats(struct impl *impl)
+{
+	uint8_t buffer[4096];
+	spa_auto(spa_pod_dynamic_builder) b = { 0 };
+	uint32_t state = 0;
+	struct spa_pod *param;
+	struct spa_pod_frame f;
+	int res;
+
+	if (!impl->recheck_format)
+		return 0;
+
+	spa_log_debug(impl->log, "updating peer formats");
+
+	spa_node_send_command(impl->follower,
+			&SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_ParamBegin));
+
+	spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
+	spa_pod_builder_push_struct(&b.b, &f);
+
+	while (true) {
+		res = node_port_enum_params_sync(impl, impl->follower,
+					impl->direction, 0,
+					SPA_PARAM_EnumFormat, &state,
+					NULL, &param, &b.b);
+		if (res != 1)
+			break;
+	}
+	param = spa_pod_builder_pop(&b.b, &f);
+
+	spa_node_send_command(impl->follower,
+			&SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_ParamEnd));
+
+	spa_pod_simplify(&b.b, &param, param);
+	spa_debug_log_pod(impl->log, SPA_LOG_LEVEL_DEBUG, 0, NULL, param);
+
+	res = spa_node_port_set_param(impl->target,
+				   SPA_DIRECTION_REVERSE(impl->direction), 0,
+				   SPA_PARAM_PeerFormats, 0, param);
+
+	impl->recheck_format = false;
+
+	spa_log_debug(impl->log, "done updating peer formats: %d", res);
+
+	return 0;
+}
+
+
 static struct spa_pod *merge_objects(struct impl *this, struct spa_pod_builder *b, uint32_t id,
 			struct spa_pod_object *o1, struct spa_pod_object *o2)
 {
@@ -862,7 +970,7 @@ static struct spa_pod *merge_objects(struct impl *this, struct spa_pod_builder *
 	struct spa_pod_builder_state state;
 	int res = 0;
 
-	if (o2 == NULL || SPA_POD_TYPE(o1) != SPA_POD_TYPE(o2))
+	if (o2 == NULL || o1->pod.type != o2->pod.type)
 		return (struct spa_pod*)o1;
 
 	spa_pod_builder_push_object(b, &f, o1->body.type, o1->body.id);
@@ -905,7 +1013,7 @@ static int negotiate_format(struct impl *this)
 	if (this->have_format && !this->recheck_format)
 		return 0;
 
-	this->recheck_format = false;
+	update_param_peer_formats(this);
 
 	spa_pod_builder_init(&b, buffer, sizeof(buffer));
 
@@ -915,11 +1023,11 @@ static int negotiate_format(struct impl *this)
 	/* The target has been negotiated on its other ports and so it can propose
 	 * a passthrough format or an ideal conversion. We use the suggestions of the
 	 * target to find the best follower format */
-	for (fstate = 0;;) {
+	for (tstate = 0;;) {
 		format = NULL;
-		res = spa_node_port_enum_params_sync(this->target,
+		res = node_port_enum_params_sync(this, this->target,
 					SPA_DIRECTION_REVERSE(this->direction), 0,
-					SPA_PARAM_EnumFormat, &fstate,
+					SPA_PARAM_EnumFormat, &tstate,
 					NULL, &format, &b);
 
 		if (res == -ENOENT)
@@ -927,14 +1035,20 @@ static int negotiate_format(struct impl *this)
 		else if (res <= 0)
 			break;
 
-		tstate = 0;
-		fres = spa_node_port_enum_params_sync(this->follower,
+		if (format != NULL)
+			spa_debug_log_pod(this->log, SPA_LOG_LEVEL_DEBUG, 0, NULL, format);
+
+		fstate = 0;
+		fres = node_port_enum_params_sync(this, this->follower,
 					this->direction, 0,
-					SPA_PARAM_EnumFormat, &tstate,
+					SPA_PARAM_EnumFormat, &fstate,
 					format, &format, &b);
 		if (fres == 0 && res == 1)
 			continue;
 
+		if (format != NULL)
+			spa_debug_log_pod(this->log, SPA_LOG_LEVEL_DEBUG, 0, NULL, format);
+
 		res = fres;
 		break;
 	}
@@ -1171,7 +1285,7 @@ static void convert_port_info(void *data,
 			port_id--;
 	} else if (info) {
 		pi = *info;
-		pi.flags = this->follower_port_flags &
+		pi.flags |= this->follower_port_flags &
 			(SPA_PORT_FLAG_LIVE |
 			 SPA_PORT_FLAG_PHYSICAL |
 			 SPA_PORT_FLAG_TERMINAL);
@@ -1189,7 +1303,7 @@ static void convert_result(void *data, int seq, int res, uint32_t type, const vo
 {
 	struct impl *this = data;
 
-	if (this->target == this->follower)
+	if (this->target == this->follower || this->in_enum_sync)
 		return;
 
 	spa_log_trace(this->log, "%p: result %d %d", this, seq, res);
@@ -1208,8 +1322,8 @@ static void follower_info(void *data, const struct spa_node_info *info)
 	struct impl *this = data;
 	uint32_t i;
 
-	spa_log_debug(this->log, "%p: info change:%08"PRIx64, this,
-			info->change_mask);
+	spa_log_debug(this->log, "%p: info change:%08"PRIx64" %d:%d", this,
+			info->change_mask, info->max_input_ports, info->max_output_ports);
 
 	if (this->follower_removing)
 		return;
@@ -1218,7 +1332,7 @@ static void follower_info(void *data, const struct spa_node_info *info)
 
 	if (info->max_input_ports > 0)
 		this->direction = SPA_DIRECTION_INPUT;
-        else
+	else
 		this->direction = SPA_DIRECTION_OUTPUT;
 
 	if (this->direction == SPA_DIRECTION_INPUT) {
@@ -1366,7 +1480,7 @@ static void follower_result(void *data, int seq, int res, uint32_t type, const v
 {
 	struct impl *this = data;
 
-	if (this->target != this->follower)
+	if (this->target != this->follower || this->in_enum_sync)
 		return;
 
 	spa_log_trace(this->log, "%p: result %d %d", this, seq, res);
@@ -1399,6 +1513,20 @@ static const struct spa_node_events follower_node_events = {
 	.event = follower_event,
 };
 
+static void follower_probe_info(void *data, const struct spa_node_info *info)
+{
+	struct impl *this = data;
+	if (info->max_input_ports > 0)
+		this->direction = SPA_DIRECTION_INPUT;
+        else
+		this->direction = SPA_DIRECTION_OUTPUT;
+}
+
+static const struct spa_node_events follower_probe_events = {
+	SPA_VERSION_NODE_EVENTS,
+	.info = follower_probe_info,
+};
+
 static int follower_ready(void *data, int status)
 {
 	struct impl *this = data;
@@ -1546,33 +1674,6 @@ impl_node_remove_port(void *object, enum spa_direction direction, uint32_t port_
 	return spa_node_remove_port(this->target, direction, port_id);
 }
 
-static int follower_port_enum_params(struct impl *this,
-				enum spa_direction direction, uint32_t port_id,
-				uint32_t id, uint32_t idx,
-				struct spa_result_node_params *result,
-				const struct spa_pod *filter,
-				struct spa_pod_builder *builder)
-{
-	int res;
-	if (result->next < 0x100000 && this->follower_params_flags[idx] & SPA_PARAM_INFO_READ) {
-		if ((res = spa_node_port_enum_params_sync(this->follower, direction, port_id,
-				id, &result->next, filter, &result->param, builder)) == 1)
-			return res;
-		result->next = 0x100000;
-	}
-	if (result->next < 0x200000 &&
-	    this->follower != this->target) {
-		result->next &= 0xfffff;
-		if ((res = spa_node_port_enum_params_sync(this->target, direction, port_id,
-				id, &result->next, filter, &result->param, builder)) == 1) {
-			result->next |= 0x100000;
-			return res;
-		}
-		result->next = 0x200000;
-	}
-	return 0;
-}
-
 static int
 impl_node_port_enum_params(void *object, int seq,
 			   enum spa_direction direction, uint32_t port_id,
@@ -1580,12 +1681,6 @@ impl_node_port_enum_params(void *object, int seq,
 			   const struct spa_pod *filter)
 {
 	struct impl *this = object;
-	uint8_t buffer[4096];
-	spa_auto(spa_pod_dynamic_builder) b = { 0 };
-	struct spa_pod_builder_state state;
-	struct spa_result_node_params result;
-	uint32_t count = 0;
-	int res;
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 	spa_return_val_if_fail(num != 0, -EINVAL);
@@ -1593,37 +1688,10 @@ impl_node_port_enum_params(void *object, int seq,
 	if (direction != this->direction)
 		port_id++;
 
-	spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
-	spa_pod_builder_get_state(&b.b, &state);
+	spa_log_debug(this->log, "%p: %d %u %u %u", this, seq, id, start, num);
 
-	result.id = id;
-	result.next = start;
-next:
-	result.index = result.next;
-
-	spa_log_debug(this->log, "%p: %d id:%u", this, seq, id);
-
-	spa_pod_builder_reset(&b.b, &state);
-
-	switch (id) {
-	case SPA_PARAM_EnumFormat:
-		res = follower_port_enum_params(this, direction, port_id,
-				id, IDX_EnumFormat, &result, filter, &b.b);
-		break;
-	default:
-		return spa_node_port_enum_params(this->target, seq, direction, port_id, id,
-			start, num, filter);
-	}
-	if (res != 1)
-		return res;
-
-	spa_node_emit_result(&this->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
-	count++;
-
-	if (count != num)
-		goto next;
-
-	return 0;
+	return spa_node_port_enum_params(this->target, seq, direction, port_id, id,
+				start, num, filter);
 }
 
 static int
@@ -1814,14 +1882,38 @@ static const struct spa_node_methods impl_node = {
 	.process = impl_node_process,
 };
 
-static int load_plugin_from(struct impl *this, const struct spa_dict *info,
-		const char *convertname, struct spa_handle **handle, struct spa_node **iface)
+static int load_converter(struct impl *this, const struct spa_dict *info,
+		const struct spa_support *support, uint32_t n_support)
 {
+	const char* factory_name = NULL;
 	struct spa_handle *hnd_convert = NULL;
 	void *iface_conv = NULL;
-	hnd_convert = spa_plugin_loader_load(this->ploader, convertname, info);
-	if (!hnd_convert)
-		return -EINVAL;
+	struct spa_dict_item *items;
+	struct spa_dict cinfo;
+	char direction[16];
+	uint32_t i;
+
+	items = alloca((info->n_items + 1) * sizeof(struct spa_dict_item));
+	cinfo = SPA_DICT(items, 0);
+	for (i = 0; i < info->n_items; i++)
+		items[cinfo.n_items++] = info->items[i];
+
+	snprintf(direction, sizeof(direction), "%s",
+			SPA_DIRECTION_REVERSE(this->direction) == SPA_DIRECTION_INPUT ?
+			"input" : "output");
+	items[cinfo.n_items++] = SPA_DICT_ITEM("convert.direction", direction);
+
+	factory_name = spa_dict_lookup(&cinfo, "video.adapt.converter");
+	if (factory_name == NULL)
+		return 0;
+
+	if (this->ploader) {
+		hnd_convert = spa_plugin_loader_load(this->ploader, factory_name, &cinfo);
+		if (!hnd_convert)
+			return -EINVAL;
+	} else {
+		return -ENOTSUP;
+	}
 
 	spa_handle_get_interface(hnd_convert, SPA_TYPE_INTERFACE_Node, &iface_conv);
 	if (iface_conv == NULL) {
@@ -1829,30 +1921,13 @@ static int load_plugin_from(struct impl *this, const struct spa_dict *info,
 		return -EINVAL;
 	}
 
-	*handle = hnd_convert;
-	*iface = iface_conv;
+	this->hnd_convert = hnd_convert;
+	this->convert = iface_conv;
+	this->convertname = strdup(factory_name);
 
 	return 0;
 }
 
-static int load_converter(struct impl *this, const struct spa_dict *info)
-{
-	int ret;
-	if (!this->ploader || !info)
-		return -EINVAL;
-
-	const char* factory_name = spa_dict_lookup(info, "video.adapt.converter");
-
-	if (factory_name) {
-		ret = load_plugin_from(this, info, factory_name, &this->hnd_convert, &this->convert);
-		if (ret >= 0) {
-			this->convertname = strdup(factory_name);
-			return ret;
-		}
-	}
-	return 0;
-}
-
 static int impl_get_interface(struct spa_handle *handle, const char *type, void **interface)
 {
 	struct impl *this;
@@ -1910,6 +1985,7 @@ impl_init(const struct spa_handle_factory *factory,
 	struct impl *this;
 	const char *str;
 	int ret;
+	struct spa_hook probe_listener;
 
 	spa_return_val_if_fail(factory != NULL, -EINVAL);
 	spa_return_val_if_fail(handle != NULL, -EINVAL);
@@ -1944,7 +2020,12 @@ impl_init(const struct spa_handle_factory *factory,
 			SPA_VERSION_NODE,
 			&impl_node, this);
 
-	ret = load_converter(this, info);
+	/* just probe the ports to get the direction */
+	spa_zero(probe_listener);
+	spa_node_add_listener(this->follower, &probe_listener, &follower_probe_events, this);
+	spa_hook_remove(&probe_listener);
+
+	ret = load_converter(this, info, support, n_support);
 	spa_log_info(this->log, "%p: loaded converter %s, hnd %p, convert %p", this,
 			this->convertname, this->hnd_convert, this->convert);
 	if (ret < 0)
@@ -1952,10 +2033,11 @@ impl_init(const struct spa_handle_factory *factory,
 
 	if (this->convert == NULL) {
 		this->target = this->follower;
-		this->passthrough = true;
+		this->mode = SPA_PARAM_PORT_CONFIG_MODE_passthrough;
 	} else {
 		this->target = this->convert;
-		this->passthrough = false;
+		/* the actual mode is selected below */
+		this->mode = SPA_PARAM_PORT_CONFIG_MODE_none;
 	}
 
 	this->info_all = SPA_NODE_CHANGE_MASK_FLAGS |
@@ -1963,8 +2045,7 @@ impl_init(const struct spa_handle_factory *factory,
 		SPA_NODE_CHANGE_MASK_PARAMS;
 	this->info = SPA_NODE_INFO_INIT();
 	this->info.flags = SPA_NODE_FLAG_RT |
-		0;
-		//SPA_NODE_FLAG_NEED_CONFIGURE;
+		SPA_NODE_FLAG_NEED_CONFIGURE;
 	this->params[IDX_EnumFormat] = SPA_PARAM_INFO(SPA_PARAM_EnumFormat, SPA_PARAM_INFO_READ);
 	this->params[IDX_PropInfo] = SPA_PARAM_INFO(SPA_PARAM_PropInfo, SPA_PARAM_INFO_READ);
 	this->params[IDX_Props] = SPA_PARAM_INFO(SPA_PARAM_Props, SPA_PARAM_INFO_READWRITE);
@@ -1981,16 +2062,18 @@ impl_init(const struct spa_handle_factory *factory,
 			&this->follower_listener, &follower_node_events, this);
 	spa_node_set_callbacks(this->follower, &follower_node_callbacks, this);
 
+	if (this->convert != NULL)
+		update_param_peer_formats(this);
+
 	// TODO: adapt port bootstrap for arbitrary converter (incl. dummy)
 	if (this->convert) {
 		spa_node_add_listener(this->convert,
 				&this->convert_listener, &convert_node_events, this);
 
 		if (strcmp(this->convertname, "video.convert.dummy") == 0) {
-			configure_convert(this, SPA_PARAM_PORT_CONFIG_MODE_none);
 			reconfigure_mode(this, SPA_PARAM_PORT_CONFIG_MODE_passthrough, this->direction, NULL);
 		} else {
-			configure_convert(this, SPA_PARAM_PORT_CONFIG_MODE_convert);
+			reconfigure_mode(this, SPA_PARAM_PORT_CONFIG_MODE_convert, this->direction, NULL);
 		}
 	} else {
 		reconfigure_mode(this, SPA_PARAM_PORT_CONFIG_MODE_passthrough, this->direction, NULL);
