@@ -25,6 +25,7 @@
 #include <unistd.h>
 
 #include <spa/param/video/format.h>
+#include <spa/pod/iter.h>
 #include <spa/pod/builder.h>
 #include <spa/utils/result.h>
 
@@ -46,7 +47,9 @@ GST_DEBUG_CATEGORY_STATIC (pipewire_src_debug);
 #define DEFAULT_RESEND_LAST     false
 #define DEFAULT_KEEPALIVE_TIME  0
 #define DEFAULT_AUTOCONNECT     true
-#define DEFAULT_USE_BUFFERPOOL USE_BUFFERPOOL_AUTO
+#define DEFAULT_USE_BUFFERPOOL  USE_BUFFERPOOL_AUTO
+#define DEFAULT_ON_DISCONNECT   GST_PIPEWIRE_SRC_ON_DISCONNECT_NONE
+#define DEFAULT_PROVIDE_CLOCK   TRUE
 
 enum
 {
@@ -64,8 +67,29 @@ enum
   PROP_KEEPALIVE_TIME,
   PROP_AUTOCONNECT,
   PROP_USE_BUFFERPOOL,
+  PROP_ON_DISCONNECT,
+  PROP_PROVIDE_CLOCK,
 };
 
+GType
+gst_pipewire_src_on_disconnect_get_type (void)
+{
+  static gsize on_disconnect_type = 0;
+  static const GEnumValue on_disconnect[] = {
+    {GST_PIPEWIRE_SRC_ON_DISCONNECT_NONE, "GST_PIPEWIRE_SRC_ON_DISCONNECT_NONE", "none"},
+    {GST_PIPEWIRE_SRC_ON_DISCONNECT_EOS, "GST_PIPEWIRE_SRC_ON_DISCONNECT_EOS", "eos"},
+    {GST_PIPEWIRE_SRC_ON_DISCONNECT_ERROR, "GST_PIPEWIRE_SRC_ON_DISCONNECT_ERROR", "error"},
+    {0, NULL, NULL},
+  };
+
+  if (g_once_init_enter (&on_disconnect_type)) {
+    GType tmp =
+        g_enum_register_static ("GstPipeWireSrcOnDisconnect", on_disconnect);
+    g_once_init_leave (&on_disconnect_type, tmp);
+  }
+
+  return (GType) on_disconnect_type;
+}
 
 static GstStaticPadTemplate gst_pipewire_src_template =
 GST_STATIC_PAD_TEMPLATE ("src",
@@ -170,6 +194,20 @@ gst_pipewire_src_set_property (GObject * object, guint prop_id,
         pwsrc->use_bufferpool = USE_BUFFERPOOL_NO;
       break;
 
+    case PROP_ON_DISCONNECT:
+      pwsrc->on_disconnect = g_value_get_enum (value);
+      break;
+
+    case PROP_PROVIDE_CLOCK:
+      gboolean provide = g_value_get_boolean (value);
+      GST_OBJECT_LOCK (pwsrc);
+      if (provide)
+        GST_OBJECT_FLAG_SET (pwsrc, GST_ELEMENT_FLAG_PROVIDE_CLOCK);
+      else
+        GST_OBJECT_FLAG_UNSET (pwsrc, GST_ELEMENT_FLAG_PROVIDE_CLOCK);
+      GST_OBJECT_UNLOCK (pwsrc);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -235,6 +273,18 @@ gst_pipewire_src_get_property (GObject * object, guint prop_id,
       g_value_set_boolean (value, !!pwsrc->use_bufferpool);
       break;
 
+    case PROP_ON_DISCONNECT:
+      g_value_set_enum (value, pwsrc->on_disconnect);
+      break;
+
+    case PROP_PROVIDE_CLOCK:
+      gboolean result;
+      GST_OBJECT_LOCK (pwsrc);
+      result = GST_OBJECT_FLAG_IS_SET (pwsrc, GST_ELEMENT_FLAG_PROVIDE_CLOCK);
+      GST_OBJECT_UNLOCK (pwsrc);
+      g_value_set_boolean (value, result);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -414,6 +464,25 @@ gst_pipewire_src_class_init (GstPipeWireSrcClass * klass)
                                                          G_PARAM_READWRITE |
                                                          G_PARAM_STATIC_STRINGS));
 
+   g_object_class_install_property (gobject_class,
+                                    PROP_ON_DISCONNECT,
+                                    g_param_spec_enum ("on-disconnect",
+                                                       "On disconnect",
+                                                       "Action to take on disconnect",
+                                                        GST_TYPE_PIPEWIRE_SRC_ON_DISCONNECT,
+                                                        DEFAULT_ON_DISCONNECT,
+                                                        G_PARAM_READWRITE |
+                                                        G_PARAM_STATIC_STRINGS));
+
+  g_object_class_install_property (gobject_class,
+                                   PROP_PROVIDE_CLOCK,
+                                   g_param_spec_boolean ("provide-clock",
+                                                         "Provide Clock",
+                                                         "Provide a clock to be used as the global pipeline clock",
+                                                         DEFAULT_PROVIDE_CLOCK,
+                                                         G_PARAM_READWRITE |
+                                                         G_PARAM_STATIC_STRINGS));
+
   gstelement_class->provide_clock = gst_pipewire_src_provide_clock;
   gstelement_class->change_state = gst_pipewire_src_change_state;
   gstelement_class->send_event = gst_pipewire_src_send_event;
@@ -462,6 +531,9 @@ gst_pipewire_src_init (GstPipeWireSrc * src)
   src->autoconnect = DEFAULT_AUTOCONNECT;
   src->min_latency = 0;
   src->max_latency = GST_CLOCK_TIME_NONE;
+  src->n_buffers = 0;
+  src->flushing_on_remove_buffer = FALSE;
+  src->on_disconnect = DEFAULT_ON_DISCONNECT;
 
   src->transform_value = UINT32_MAX;
 }
@@ -469,11 +541,26 @@ gst_pipewire_src_init (GstPipeWireSrc * src)
 static gboolean
 buffer_recycle (GstMiniObject *obj)
 {
-  GstPipeWireSrc *src;
-  GstPipeWirePoolData *data;
+  GstPipeWirePoolData *data = gst_pipewire_pool_get_data (GST_BUFFER_CAST(obj));
+  GstPipeWireSrc *src = data->owner;
   int res;
 
-  data = gst_pipewire_pool_get_data (GST_BUFFER_CAST(obj));
+  if (src->flushing_on_remove_buffer) {
+    /*
+     * If a flush-start was initiated, this might be called by elements like
+     * queues downstream purging buffers from their internal queues. This can
+     * deadlock if queues use min-threshold-buffers/bytes/time with src_create
+     * trying to take the loop lock and buffer_recycle trying to take the loop
+     * lock down below. We return from here, to prevent deadlock with streaming
+     * thread in a queue thread.
+     *
+     * We will take care of queueing the buffer in on_remove_buffer.
+     */
+    GstBuffer *buffer = GST_BUFFER_CAST(obj);
+    GST_DEBUG_OBJECT (src,
+        "flush-start initiated, skipping buffer recycle %p", buffer);
+    return TRUE;
+  }
 
   GST_OBJECT_LOCK (data->pool);
   if (!obj->dispose) {
@@ -482,7 +569,6 @@ buffer_recycle (GstMiniObject *obj)
   }
 
   GST_BUFFER_FLAGS (obj) = data->flags;
-  src = data->owner;
 
   pw_thread_loop_lock (src->stream->core->loop);
   if (!obj->dispose) {
@@ -519,6 +605,8 @@ on_add_buffer (void *_data, struct pw_buffer *b)
   data->owner = pwsrc;
   data->queued = TRUE;
   GST_MINI_OBJECT_CAST (data->buf)->dispose = buffer_recycle;
+
+  pwsrc->n_buffers++;
 }
 
 static void
@@ -527,17 +615,76 @@ on_remove_buffer (void *_data, struct pw_buffer *b)
   GstPipeWireSrc *pwsrc = _data;
   GstPipeWirePoolData *data = b->user_data;
   GstBuffer *buf = data->buf;
+  gboolean flush_on_remove;
   int res;
 
-  GST_DEBUG_OBJECT (pwsrc, "remove buffer %p", buf);
+  GST_DEBUG_OBJECT (pwsrc, "remove buffer %p, queued: %d",
+      buf, data->queued);
 
   GST_MINI_OBJECT_CAST (buf)->dispose = NULL;
 
+  flush_on_remove =
+        pwsrc->on_disconnect == GST_PIPEWIRE_SRC_ON_DISCONNECT_ERROR ||
+        pwsrc->on_disconnect == GST_PIPEWIRE_SRC_ON_DISCONNECT_EOS;
+
+  if (flush_on_remove && !pwsrc->flushing_on_remove_buffer) {
+      pwsrc->flushing_on_remove_buffer = TRUE;
+
+      GST_DEBUG_OBJECT (pwsrc, "flush-start on remove buffer");
+      /*
+       * It is possible that when buffers are being removed, a downstream
+       * element can be holding on to a buffer or in the middle of rendering
+       * the same. Former is possible with queues min-threshold-buffers or
+       * similar. Latter can result in a crash during gst_video_frame_copy.
+       *
+       * We send a flush-start event downstream to make elements discard
+       * any buffers they may be holding on to as well as return from their
+       * chain function ASAP.
+       */
+      gst_pad_push_event (GST_BASE_SRC_PAD (pwsrc),
+          gst_event_new_flush_start ());
+  }
+
   if (data->queued) {
     gst_buffer_unref (buf);
   } else {
     if ((res = pw_stream_queue_buffer (pwsrc->stream->pwstream, b)) < 0)
-      GST_WARNING_OBJECT (pwsrc, "can't queue removed buffer %p, %s", buf, spa_strerror(res));
+      GST_WARNING_OBJECT (pwsrc, "can't queue removed buffer %p, %s",
+          buf, spa_strerror(res));
+    else
+      GST_DEBUG_OBJECT (pwsrc, "queued buffer %p", buf);
+  }
+
+  pwsrc->n_buffers--;
+
+  if (pwsrc->n_buffers == 0) {
+    GST_DEBUG_OBJECT (pwsrc, "removed all buffers");
+
+    pwsrc->flushing_on_remove_buffer = FALSE;
+
+    switch (pwsrc->on_disconnect) {
+      case GST_PIPEWIRE_SRC_ON_DISCONNECT_ERROR:
+        GST_DEBUG_OBJECT (pwsrc, "flush-stop on removing all buffers");
+        gst_pad_push_event (GST_BASE_SRC_PAD (pwsrc),
+            gst_event_new_flush_stop (FALSE));
+
+        GST_ELEMENT_ERROR (pwsrc, RESOURCE, NOT_FOUND,
+            ("all buffers have been removed"),
+            ("PipeWire link to remote node was destroyed"));
+        break;
+      case GST_PIPEWIRE_SRC_ON_DISCONNECT_EOS:
+        GST_DEBUG_OBJECT (pwsrc, "flush-stop on removing all buffers");
+        gst_pad_push_event (GST_BASE_SRC_PAD (pwsrc),
+            gst_event_new_flush_stop (FALSE));
+
+        GST_DEBUG_OBJECT (pwsrc, "sending eos downstream");
+        gst_pad_push_event (GST_BASE_SRC_PAD (pwsrc),
+            gst_event_new_eos());
+        break;
+      case GST_PIPEWIRE_SRC_ON_DISCONNECT_NONE:
+        GST_DEBUG_OBJECT (pwsrc, "stream closed or removed");
+        break;
+    }
   }
 }
 
@@ -568,6 +715,7 @@ static GstBuffer *dequeue_buffer(GstPipeWireSrc *pwsrc)
   struct spa_meta_header *h;
   struct spa_meta_region *crop;
   enum spa_meta_videotransform_value transform_value;
+  struct spa_meta_cursor *cursor;
   struct pw_time time;
   guint i;
 
@@ -660,9 +808,19 @@ static GstBuffer *dequeue_buffer(GstPipeWireSrc *pwsrc)
     pwsrc->transform_value = transform_value;
   }
 
-  if (pwsrc->is_video) {
-    gsize video_size = 0;
+  cursor = data->cursor;
+  if (cursor && cursor->id != 0) {
+    /* TODO: at some point, maybe we can figure out width and height from the bitmap,
+     * and even add that to the meta itself */
+    gst_buffer_add_video_region_of_interest_meta (buf, "cursor", cursor->position.x, cursor->position.y, 0, 0);
+  }
+
+  if (pwsrc->is_rawvideo) {
     GstVideoInfo *info = &pwsrc->video_info;
+    uint32_t n_datas = b->buffer->n_datas;
+    uint32_t n_planes = GST_VIDEO_INFO_N_PLANES (info);
+    gsize video_size = 0;
+
     GstVideoMeta *meta = gst_buffer_add_video_meta_full (buf, GST_VIDEO_FRAME_FLAG_NONE,
                              GST_VIDEO_INFO_FORMAT (info),
                              GST_VIDEO_INFO_WIDTH (info),
@@ -671,8 +829,10 @@ static GstBuffer *dequeue_buffer(GstPipeWireSrc *pwsrc)
                              info->offset,
                              info->stride);
 
-    for (i = 0; i < MIN (b->buffer->n_datas, GST_VIDEO_MAX_PLANES); i++) {
+    for (i = 0; i < MIN (n_datas, n_planes); i++) {
       struct spa_data *d = &b->buffer->datas[i];
+      /* don't add the chunk offset here, this is done below when we
+       * share/copy the memory in the target buffer below */
       meta->offset[i] = video_size;
       meta->stride[i] = d->chunk->stride;
 
@@ -680,6 +840,10 @@ static GstBuffer *dequeue_buffer(GstPipeWireSrc *pwsrc)
     }
   }
 
+  if (b->buffer->n_datas != gst_buffer_n_memory(data->buf)) {
+      GST_ERROR_OBJECT(pwsrc, "n_datas != n_memory, (%d != %d)", b->buffer->n_datas, gst_buffer_n_memory(data->buf));
+  }
+
   for (i = 0; i < b->buffer->n_datas; i++) {
     struct spa_data *d = &b->buffer->datas[i];
 
@@ -730,13 +894,29 @@ on_state_changed (void *data,
                   enum pw_stream_state state, const char *error)
 {
   GstPipeWireSrc *pwsrc = data;
+  GstState current_state = GST_ELEMENT_CAST (pwsrc)->current_state;
 
-  GST_DEBUG ("got stream state %s", pw_stream_state_as_string (state));
+  GST_DEBUG_OBJECT (pwsrc, "got stream state %s", pw_stream_state_as_string (state));
 
   switch (state) {
     case PW_STREAM_STATE_UNCONNECTED:
     case PW_STREAM_STATE_CONNECTING:
+      break;
     case PW_STREAM_STATE_PAUSED:
+      /*
+       * We may see a driver/quantum/clock rate change on switching audio
+       * sources. The same is not applicable for video.
+       *
+       * We post the clock lost message here to take care of a possible
+       * jump or shift in base_time/clock for the pipeline. Application
+       * must handle the clock lost message in it's bus handler by pausing
+       * the pipeline and then setting it back to playing.
+       */
+      if (current_state == GST_STATE_PLAYING && !pwsrc->is_video)
+        gst_element_post_message (GST_ELEMENT_CAST (pwsrc),
+            gst_message_new_clock_lost (GST_OBJECT_CAST (pwsrc),
+                GST_CLOCK_CAST (pwsrc->stream->clock)));
+      break;
     case PW_STREAM_STATE_STREAMING:
       break;
     case PW_STREAM_STATE_ERROR:
@@ -982,7 +1162,7 @@ gst_pipewire_src_negotiate (GstBaseSrc * basesrc)
   GST_DEBUG_OBJECT (basesrc, "connect capture with path %s, target-object %s",
                     pwsrc->stream->path, pwsrc->stream->target_object);
 
-  pwsrc->possible_caps = possible_caps;
+  gst_caps_replace (&pwsrc->possible_caps, possible_caps);
   pwsrc->negotiated = FALSE;
 
   enum pw_stream_flags flags;
@@ -1019,7 +1199,6 @@ gst_pipewire_src_negotiate (GstBaseSrc * basesrc)
   }
 
   negotiated_caps = g_steal_pointer (&pwsrc->caps);
-  pwsrc->possible_caps = NULL;
   pw_thread_loop_unlock (pwsrc->stream->core->loop);
 
   if (negotiated_caps == NULL)
@@ -1027,7 +1206,7 @@ gst_pipewire_src_negotiate (GstBaseSrc * basesrc)
 
   gst_pipewire_clock_reset (GST_PIPEWIRE_CLOCK (pwsrc->stream->clock), 0);
 
-  GST_DEBUG_OBJECT (pwsrc, "set format %" GST_PTR_FORMAT, negotiated_caps);
+  GST_INFO_OBJECT (pwsrc, "set format %" GST_PTR_FORMAT, negotiated_caps);
   result = gst_base_src_set_caps (GST_BASE_SRC (pwsrc), negotiated_caps);
   if (!result)
     goto no_caps;
@@ -1140,10 +1319,21 @@ handle_format_change (GstPipeWireSrc *pwsrc,
           pw_stream_set_error (pwsrc->stream->pwstream, -EINVAL, "internal error");
           return;
         }
+        pwsrc->is_rawvideo = TRUE;
       } else {
         gst_video_info_dma_drm_init (&pwsrc->drm_info);
 #endif
         gst_video_info_from_caps (&pwsrc->video_info, pwsrc->caps);
+
+        if (GST_VIDEO_FORMAT_INFO_IS_VALID_RAW (pwsrc->video_info.finfo)
+#ifdef HAVE_GSTREAMER_DMA_DRM
+            && GST_VIDEO_FORMAT_INFO_FORMAT (pwsrc->video_info.finfo) != GST_VIDEO_FORMAT_DMA_DRM
+#endif
+            )
+          pwsrc->is_rawvideo = TRUE;
+        else
+          pwsrc->is_rawvideo = FALSE;
+
 #ifdef HAVE_GSTREAMER_DMA_DRM
       }
 #endif
@@ -1156,14 +1346,15 @@ handle_format_change (GstPipeWireSrc *pwsrc,
   } else {
     pwsrc->negotiated = FALSE;
     pwsrc->is_video = FALSE;
+    pwsrc->is_rawvideo = FALSE;
   }
 
   if (pwsrc->caps) {
-    const struct spa_pod *params[4];
+    const struct spa_pod *params[10];
     struct spa_pod_builder b = { NULL };
-    uint8_t buffer[512];
+    uint8_t buffer[16384];
     uint32_t buffers = CLAMP (16, pwsrc->min_buffers, pwsrc->max_buffers);
-    int buffertypes;
+    int buffertypes, n_params = 0;
 
     buffertypes = (1<<SPA_DATA_DmaBuf);
     if (!spa_pod_find_prop (param, NULL, SPA_FORMAT_VIDEO_modifier)) {
@@ -1173,7 +1364,7 @@ handle_format_change (GstPipeWireSrc *pwsrc,
     GST_DEBUG_OBJECT (pwsrc, "we got format %" GST_PTR_FORMAT, pwsrc->caps);
 
     spa_pod_builder_init (&b, buffer, sizeof (buffer));
-    params[0] = spa_pod_builder_add_object (&b,
+    params[n_params++] = spa_pod_builder_add_object (&b,
         SPA_TYPE_OBJECT_ParamBuffers, SPA_PARAM_Buffers,
         SPA_PARAM_BUFFERS_buffers, SPA_POD_CHOICE_RANGE_Int(buffers,
                                                             pwsrc->min_buffers,
@@ -1183,21 +1374,31 @@ handle_format_change (GstPipeWireSrc *pwsrc,
         SPA_PARAM_BUFFERS_stride,  SPA_POD_CHOICE_RANGE_Int(0, 0, INT32_MAX),
         SPA_PARAM_BUFFERS_dataType, SPA_POD_CHOICE_FLAGS_Int(buffertypes));
 
-    params[1] = spa_pod_builder_add_object (&b,
+    params[n_params++] = spa_pod_builder_add_object (&b,
         SPA_TYPE_OBJECT_ParamMeta, SPA_PARAM_Meta,
         SPA_PARAM_META_type, SPA_POD_Id(SPA_META_Header),
         SPA_PARAM_META_size, SPA_POD_Int(sizeof (struct spa_meta_header)));
-    params[2] = spa_pod_builder_add_object (&b,
+    params[n_params++] = spa_pod_builder_add_object (&b,
         SPA_TYPE_OBJECT_ParamMeta, SPA_PARAM_Meta,
         SPA_PARAM_META_type, SPA_POD_Id(SPA_META_VideoCrop),
         SPA_PARAM_META_size, SPA_POD_Int(sizeof (struct spa_meta_region)));
-    params[3] = spa_pod_builder_add_object (&b,
+    params[n_params++] = spa_pod_builder_add_object (&b,
         SPA_TYPE_OBJECT_ParamMeta, SPA_PARAM_Meta,
         SPA_PARAM_META_type, SPA_POD_Id(SPA_META_VideoTransform),
         SPA_PARAM_META_size, SPA_POD_Int(sizeof (struct spa_meta_videotransform)));
+#define CURSOR_META_SIZE(width, height) \
+        (sizeof (struct spa_meta_cursor) + \
+         sizeof (struct spa_meta_bitmap) + width * height * 4)
+    params[n_params++] = spa_pod_builder_add_object (&b,
+        SPA_TYPE_OBJECT_ParamMeta, SPA_PARAM_Meta,
+        SPA_PARAM_META_type, SPA_POD_Id(SPA_META_Cursor),
+        SPA_PARAM_META_size,
+        SPA_POD_CHOICE_RANGE_Int (CURSOR_META_SIZE(384, 384),
+          sizeof (struct spa_meta_cursor),
+          CURSOR_META_SIZE(384, 384)));
 
     GST_DEBUG_OBJECT (pwsrc, "doing finish format");
-    pw_stream_update_params (pwsrc->stream->pwstream, params, SPA_N_ELEMENTS(params));
+    pw_stream_update_params (pwsrc->stream->pwstream, params, n_params);
   } else {
     GST_WARNING_OBJECT (pwsrc, "finish format with error");
     pw_stream_set_error (pwsrc->stream->pwstream, -EINVAL, "unhandled format");
@@ -1330,6 +1531,8 @@ gst_pipewire_src_create (GstPushSrc * psrc, GstBuffer ** buffer)
   GstBuffer *buf;
   gboolean update_time = FALSE, timeout = FALSE;
   GstCaps *caps = NULL;
+  struct timespec abstime = { 0, };
+  bool have_abstime = false;
 
   pwsrc = GST_PIPEWIRE_SRC (psrc);
 
@@ -1373,13 +1576,11 @@ gst_pipewire_src_create (GstPushSrc * psrc, GstBuffer ** buffer)
       update_time = TRUE;
       GST_LOG_OBJECT (pwsrc, "EOS, send last buffer");
       break;
-    } else if (timeout) {
-      if (pwsrc->last_buffer != NULL) {
-        update_time = TRUE;
-        buf = gst_buffer_ref(pwsrc->last_buffer);
-        GST_LOG_OBJECT (pwsrc, "timeout, send keepalive buffer");
-        break;
-      }
+    } else if (timeout && pwsrc->last_buffer != NULL) {
+      update_time = TRUE;
+      buf = gst_buffer_ref(pwsrc->last_buffer);
+      GST_LOG_OBJECT (pwsrc, "timeout, send keepalive buffer");
+      break;
     } else {
       buf = dequeue_buffer (pwsrc);
       GST_LOG_OBJECT (pwsrc, "popped buffer %p", buf);
@@ -1391,9 +1592,13 @@ gst_pipewire_src_create (GstPushSrc * psrc, GstBuffer ** buffer)
     }
     timeout = FALSE;
     if (pwsrc->keepalive_time > 0) {
-      struct timespec abstime;
-      pw_thread_loop_get_time(pwsrc->stream->core->loop, &abstime,
-                      pwsrc->keepalive_time * SPA_NSEC_PER_MSEC);
+      if (!have_abstime) {
+        /* Record the time we want to timeout at once, for this loop -- the loop might get unrelated signal()s,
+        * and we don't want the keepalive time to get reset by that */
+        pw_thread_loop_get_time(pwsrc->stream->core->loop, &abstime,
+            pwsrc->keepalive_time * SPA_NSEC_PER_MSEC);
+        have_abstime = TRUE;
+      }
       if (pw_thread_loop_timed_wait_full (pwsrc->stream->core->loop, &abstime) == -ETIMEDOUT)
         timeout = TRUE;
     } else {
@@ -1464,6 +1669,7 @@ gst_pipewire_src_stop (GstBaseSrc * basesrc)
   pwsrc->eos = false;
   gst_buffer_replace (&pwsrc->last_buffer, NULL);
   gst_caps_replace(&pwsrc->caps, NULL);
+  gst_caps_replace(&pwsrc->possible_caps, NULL);
   pwsrc->transform_value = UINT32_MAX;
   pw_thread_loop_unlock (pwsrc->stream->core->loop);
 
