@@ -7,6 +7,7 @@
 #include <spa/param/audio/format.h>
 #include <spa/param/audio/format-utils.h>
 #include <spa/param/audio/raw.h>
+#include <spa/param/audio/raw-json.h>
 #include <spa/utils/json.h>
 
 #include "format.h"
@@ -130,22 +131,12 @@ uint32_t format_pa2id(enum sample_format format)
 
 const char *format_id2name(uint32_t format)
 {
-	int i;
-	for (i = 0; spa_type_audio_format[i].name; i++) {
-		if (spa_type_audio_format[i].type == format)
-			return spa_debug_type_short_name(spa_type_audio_format[i].name);
-	}
-	return "UNKNOWN";
+	return spa_type_audio_format_to_short_name(format);
 }
 
 uint32_t format_name2id(const char *name)
 {
-	int i;
-	for (i = 0; spa_type_audio_format[i].name; i++) {
-		if (spa_streq(name, spa_debug_type_short_name(spa_type_audio_format[i].name)))
-			return spa_type_audio_format[i].type;
-	}
-	return SPA_AUDIO_FORMAT_UNKNOWN;
+	return spa_type_audio_format_from_short_name(name);
 }
 
 uint32_t format_paname2id(const char *name, size_t size)
@@ -307,22 +298,12 @@ uint32_t channel_pa2id(enum channel_position channel)
 
 const char *channel_id2name(uint32_t channel)
 {
-	int i;
-	for (i = 0; spa_type_audio_channel[i].name; i++) {
-		if (spa_type_audio_channel[i].type == channel)
-			return spa_debug_type_short_name(spa_type_audio_channel[i].name);
-	}
-	return "UNK";
+	return spa_type_audio_channel_to_short_name(channel);
 }
 
 uint32_t channel_name2id(const char *name)
 {
-	int i;
-	for (i = 0; spa_type_audio_channel[i].name; i++) {
-		if (strcmp(name, spa_debug_type_short_name(spa_type_audio_channel[i].name)) == 0)
-			return spa_type_audio_channel[i].type;
-	}
-	return SPA_AUDIO_CHANNEL_UNKNOWN;
+	return spa_type_audio_channel_from_short_name(name);
 }
 
 enum channel_position channel_id2pa(uint32_t id, uint32_t *aux)
@@ -367,11 +348,20 @@ uint32_t channel_paname2id(const char *name, size_t size)
 
 void channel_map_to_positions(const struct channel_map *map, uint32_t *pos)
 {
-	int i;
-	for (i = 0; i < map->channels; i++)
+	uint32_t i, channels = SPA_MIN(map->channels, SPA_AUDIO_MAX_CHANNELS);
+	for (i = 0; i < channels; i++)
 		pos[i] = map->map[i];
 }
 
+void positions_to_channel_map(const uint32_t *pos, uint32_t channels, struct channel_map *map)
+{
+	uint32_t i;
+	channels = SPA_MIN(channels, CHANNELS_MAX);
+	for (i = 0; i < channels; i++)
+		map->map[i] = pos[i];
+	map->channels = channels;
+}
+
 void channel_map_parse(const char *str, struct channel_map *map)
 {
 	const char *p = str;
@@ -441,7 +431,7 @@ void channel_map_parse(const char *str, struct channel_map *map)
 		};
 	} else {
 		channels = map->channels = 0;
-		while (*p && channels < SPA_AUDIO_MAX_CHANNELS) {
+		while (*p && channels < CHANNELS_MAX) {
 			uint32_t chname;
 
 			if ((len = strcspn(p, ",")) == 0)
@@ -458,18 +448,9 @@ void channel_map_parse(const char *str, struct channel_map *map)
 
 void channel_map_parse_position(const char *str, struct channel_map *map)
 {
-	struct spa_json it[2];
-	char v[256];
-
-	spa_json_init(&it[0], str, strlen(str));
-	if (spa_json_enter_array(&it[0], &it[1]) <= 0)
-		spa_json_init(&it[1], str, strlen(str));
-
-	map->channels = 0;
-	while (spa_json_get_string(&it[1], v, sizeof(v)) > 0 &&
-	    map->channels < SPA_AUDIO_MAX_CHANNELS) {
-		map->map[map->channels++] = channel_name2id(v);
-	}
+	uint32_t channels = 0, position[SPA_AUDIO_MAX_CHANNELS];
+	spa_audio_parse_position(str, strlen(str), position, &channels);
+	positions_to_channel_map(position, channels, map);
 }
 
 bool channel_map_valid(const struct channel_map *map)
@@ -596,11 +577,11 @@ int format_parse_param(const struct spa_pod *param, bool collect,
 		if (info.info.raw.rate)
 		        ss->rate = info.info.raw.rate;
 		if (info.info.raw.channels)
-		        ss->channels = info.info.raw.channels;
+		        ss->channels = SPA_MIN(info.info.raw.channels, CHANNELS_MAX);
 	}
 	if (map) {
 		if (info.info.raw.channels) {
-			map->channels = info.info.raw.channels;
+			map->channels = SPA_MIN(info.info.raw.channels, CHANNELS_MAX);
 			for (i = 0; i < map->channels; i++)
 				map->map[i] = info.info.raw.position[i];
 		}
@@ -674,13 +655,13 @@ int format_info_from_spec(struct format_info *info, const struct sample_spec *ss
 	pw_properties_setf(info->props, "format.channels", "%d", ss->channels);
 	if (map && map->channels == ss->channels) {
 		char chmap[1024] = "";
-		int i, o, r;
-		uint32_t aux = 0;
+		int r;
+		uint32_t aux = 0, i, o;
 
 		for (i = 0, o = 0; i < map->channels; i++) {
 			r = snprintf(chmap+o, sizeof(chmap)-o, "%s%s", i == 0 ? "" : ",",
 					channel_id2paname(map->map[i], &aux));
-			if (r < 0 || o + r >= (int)sizeof(chmap))
+			if (r < 0 || o + r >= sizeof(chmap))
 				return -ENOSPC;
 			o += r;
 		}
@@ -702,7 +683,7 @@ static int add_int(struct format_info *info, const char *k, struct spa_pod *para
 		return -ENOENT;
 
 	val = spa_pod_get_values(&prop->value, &n_values, &choice);
-	if (val->type != SPA_TYPE_Int)
+	if (!spa_pod_is_int(val))
 		return -ENOTSUP;
 
 	if (n_values == 0)
@@ -786,9 +767,7 @@ static int format_info_iec958_from_param(struct format_info *info, struct spa_po
 	if ((info->props = pw_properties_new(NULL, NULL)) == NULL)
 		return -errno;
 
-	add_int(info, "format.rate", param, SPA_FORMAT_AUDIO_rate);
-
-	return 0;
+	return add_int(info, "format.rate", param, SPA_FORMAT_AUDIO_rate);
 }
 
 int format_info_from_param(struct format_info *info, struct spa_pod *param, uint32_t index)
@@ -824,8 +803,7 @@ static uint32_t format_info_get_format(const struct format_info *info)
 	if ((str = pw_properties_get(info->props, "format.sample_format")) == NULL)
 		return SPA_AUDIO_FORMAT_UNKNOWN;
 
-	spa_json_init(&it[0], str, strlen(str));
-	if ((len = spa_json_next(&it[0], &val)) <= 0)
+	if ((len = spa_json_begin(&it[0], str, strlen(str), &val)) <= 0)
 		return SPA_AUDIO_FORMAT_UNKNOWN;
 
 	if (spa_json_is_string(val, len))
@@ -843,8 +821,7 @@ static int format_info_get_rate(const struct format_info *info)
 	if ((str = pw_properties_get(info->props, "format.rate")) == NULL)
 		return -ENOENT;
 
-	spa_json_init(&it[0], str, strlen(str));
-	if ((len = spa_json_next(&it[0], &val)) <= 0)
+	if ((len = spa_json_begin(&it[0], str, strlen(str), &val)) <= 0)
 		return -EINVAL;
 	if (spa_json_is_int(val, len)) {
 		if (spa_json_parse_int(val, len, &v) <= 0)
@@ -880,8 +857,7 @@ int format_info_to_spec(const struct format_info *info, struct sample_spec *ss,
 	if ((str = pw_properties_get(info->props, "format.channels")) == NULL)
 		return -ENOENT;
 
-	spa_json_init(&it[0], str, strlen(str));
-	if ((len = spa_json_next(&it[0], &val)) <= 0)
+	if ((len = spa_json_begin(&it[0], str, strlen(str), &val)) <= 0)
 		return -EINVAL;
 	if (spa_json_is_float(val, len)) {
 		if (spa_json_parse_float(val, len, &f) <= 0)
@@ -895,8 +871,7 @@ int format_info_to_spec(const struct format_info *info, struct sample_spec *ss,
 		return -ENOTSUP;
 
 	if ((str = pw_properties_get(info->props, "format.channel_map")) != NULL) {
-		spa_json_init(&it[0], str, strlen(str));
-		if ((len = spa_json_next(&it[0], &val)) <= 0)
+		if ((len = spa_json_begin(&it[0], str, strlen(str), &val)) <= 0)
 			return -EINVAL;
 		if (!spa_json_is_string(val, len))
 			return -EINVAL;
