@@ -25,9 +25,10 @@ extern "C" {
 
 #define PW_BUFFERS_FLAG_NONE		0
 #define PW_BUFFERS_FLAG_NO_MEM		(1<<0)	/**< don't allocate buffer memory */
-#define PW_BUFFERS_FLAG_SHARED		(1<<1)	/**< buffers can be shared */
+#define PW_BUFFERS_FLAG_SHARED		(1<<1)	/**< buffers can be shared, This makes sure that
+						  *  the data types on the buffers don't contain
+						  *  any MemPtr, only fd based memory. */
 #define PW_BUFFERS_FLAG_DYNAMIC		(1<<2)	/**< buffers have dynamic data */
-#define PW_BUFFERS_FLAG_SHARED_MEM	(1<<3)	/**< buffers need shared memory */
 #define PW_BUFFERS_FLAG_IN_PRIORITY	(1<<4)	/**< input parameters have priority */
 #define PW_BUFFERS_FLAG_ASYNC		(1<<5)	/**< one of the nodes is async */
 
