@@ -2,6 +2,8 @@
 /* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
 /* SPDX-License-Identifier: MIT */
 
+#include "config.h"
+
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -11,8 +13,6 @@
 #include <malloc.h>
 #include <limits.h>
 
-#include "config.h"
-
 #include <spa/support/system.h>
 #include <spa/pod/parser.h>
 #include <spa/pod/filter.h>
@@ -55,6 +55,27 @@ struct impl {
 	char *sync_group;
 };
 
+static const char * const global_keys[] = {
+	PW_KEY_OBJECT_PATH,
+	PW_KEY_MODULE_ID,
+	PW_KEY_FACTORY_ID,
+	PW_KEY_CLIENT_ID,
+	PW_KEY_CLIENT_API,
+	PW_KEY_DEVICE_ID,
+	PW_KEY_PRIORITY_SESSION,
+	PW_KEY_PRIORITY_DRIVER,
+	PW_KEY_APP_NAME,
+	PW_KEY_NODE_DESCRIPTION,
+	PW_KEY_NODE_NAME,
+	PW_KEY_NODE_NICK,
+	PW_KEY_NODE_SESSION,
+	PW_KEY_MEDIA_CLASS,
+	PW_KEY_MEDIA_TYPE,
+	PW_KEY_MEDIA_CATEGORY,
+	PW_KEY_MEDIA_ROLE,
+	NULL
+};
+
 #define pw_node_resource(r,m,v,...)	pw_resource_call(r,struct pw_node_events,m,v,__VA_ARGS__)
 #define pw_node_resource_info(r,...)	pw_node_resource(r,info,0,__VA_ARGS__)
 #define pw_node_resource_param(r,...)	pw_node_resource(r,param,0,__VA_ARGS__)
@@ -209,7 +230,7 @@ do_node_prepare(struct spa_loop *loop, bool async, uint32_t seq,
 
 static void add_node_to_graph(struct pw_impl_node *node)
 {
-	pw_loop_invoke(node->data_loop, do_node_prepare, 1, NULL, 0, true, node);
+	pw_loop_locked(node->data_loop, do_node_prepare, 1, NULL, 0, node);
 }
 
 /* called from the node data loop and undoes the changes done in do_node_prepare.  */
@@ -252,7 +273,7 @@ do_node_unprepare(struct spa_loop *loop, bool async, uint32_t seq,
 
 static void remove_node_from_graph(struct pw_impl_node *node)
 {
-	pw_loop_invoke(node->data_loop, do_node_unprepare, 1, NULL, 0, true, node);
+	pw_loop_locked(node->data_loop, do_node_unprepare, 1, NULL, 0, node);
 }
 
 static void node_deactivate(struct pw_impl_node *this)
@@ -361,6 +382,8 @@ static void emit_info_changed(struct pw_impl_node *node, bool flags_changed)
 
 	if (node->global && node->info.change_mask != 0) {
 		struct pw_resource *resource;
+		if (node->info.change_mask & PW_NODE_CHANGE_MASK_PROPS)
+			pw_global_update_keys(node->global, node->info.props, global_keys);
 		spa_list_for_each(resource, &node->global->resource_list, link)
 			pw_node_resource_info(resource, &node->info);
 	}
@@ -814,8 +837,8 @@ int pw_impl_node_set_io(struct pw_impl_node *this, uint32_t id, void *data, size
 		if (data != NULL && size < sizeof(struct spa_io_position))
 			return -EINVAL;
 		pw_log_debug("%p: set position %p", this, data);
-		pw_loop_invoke(this->data_loop,
-				do_update_position, SPA_ID_INVALID, &data, sizeof(void*), true, this);
+		pw_loop_locked(this->data_loop,
+				do_update_position, SPA_ID_INVALID, &data, sizeof(void*), this);
 		break;
 	case SPA_IO_Clock:
 		if (data != NULL && size < sizeof(struct spa_io_clock))
@@ -873,8 +896,8 @@ do_add_target(struct spa_loop *loop,
 SPA_EXPORT
 int pw_impl_node_add_target(struct pw_impl_node *node, struct pw_node_target *t)
 {
-	pw_loop_invoke(node->data_loop,
-			do_add_target, SPA_ID_INVALID, &node, sizeof(void *), true, t);
+	pw_loop_locked(node->data_loop,
+			do_add_target, SPA_ID_INVALID, &node, sizeof(void *), t);
 	if (t->node)
 		pw_impl_node_emit_peer_added(node, t->node);
 
@@ -909,8 +932,8 @@ int pw_impl_node_remove_target(struct pw_impl_node *node, struct pw_node_target
 {
 	/* we also update the target list for remote nodes so that the profiler
 	 * can inspect the nodes as well */
-	pw_loop_invoke(node->data_loop,
-			do_remove_target, SPA_ID_INVALID, &node, sizeof(void *), true, t);
+	pw_loop_locked(node->data_loop,
+			do_remove_target, SPA_ID_INVALID, &node, sizeof(void *), t);
 	if (t->node)
 		pw_impl_node_emit_peer_removed(node, t->node);
 
@@ -933,27 +956,6 @@ SPA_EXPORT
 int pw_impl_node_register(struct pw_impl_node *this,
 		     struct pw_properties *properties)
 {
-	static const char * const keys[] = {
-		PW_KEY_OBJECT_PATH,
-		PW_KEY_MODULE_ID,
-		PW_KEY_FACTORY_ID,
-		PW_KEY_CLIENT_ID,
-		PW_KEY_CLIENT_API,
-		PW_KEY_DEVICE_ID,
-		PW_KEY_PRIORITY_SESSION,
-		PW_KEY_PRIORITY_DRIVER,
-		PW_KEY_APP_NAME,
-		PW_KEY_NODE_DESCRIPTION,
-		PW_KEY_NODE_NAME,
-		PW_KEY_NODE_NICK,
-		PW_KEY_NODE_SESSION,
-		PW_KEY_MEDIA_CLASS,
-		PW_KEY_MEDIA_TYPE,
-		PW_KEY_MEDIA_CATEGORY,
-		PW_KEY_MEDIA_ROLE,
-		NULL
-	};
-
 	struct pw_context *context = this->context;
 	struct pw_impl_port *port;
 
@@ -987,9 +989,8 @@ int pw_impl_node_register(struct pw_impl_node *this,
 	pw_properties_setf(this->properties, PW_KEY_OBJECT_ID, "%d", this->global->id);
 	pw_properties_setf(this->properties, PW_KEY_OBJECT_SERIAL, "%"PRIu64,
 			pw_global_get_serial(this->global));
-	this->info.props = &this->properties->dict;
 
-	pw_global_update_keys(this->global, &this->properties->dict, keys);
+	pw_global_update_keys(this->global, &this->properties->dict, global_keys);
 
 	pw_impl_node_initialized(this);
 
@@ -1112,7 +1113,6 @@ static int execute_match(void *data, const char *location, const char *action,
 	struct pw_impl_node *this = match->node;
 	if (spa_streq(action, "update-props")) {
 		match->count += pw_properties_update_string(this->properties, val, len);
-		this->info.props = &this->properties->dict;
 	}
 	return 1;
 }
@@ -1157,6 +1157,8 @@ static void check_properties(struct pw_impl_node *node)
 	node->transport_sync = pw_properties_get_bool(node->properties, PW_KEY_NODE_TRANSPORT_SYNC, false);
 	impl->cache_params =  pw_properties_get_bool(node->properties, PW_KEY_NODE_CACHE_PARAMS, true);
 	driver = pw_properties_get_bool(node->properties, PW_KEY_NODE_DRIVER, false);
+	node->exclusive = pw_properties_get_bool(node->properties, PW_KEY_NODE_EXCLUSIVE, false);
+	node->reliable = pw_properties_get_bool(node->properties, PW_KEY_NODE_RELIABLE, false);
 
 	if (node->driver != driver) {
 		pw_log_debug("%p: driver %d -> %d", node, node->driver, driver);
@@ -1308,6 +1310,10 @@ static void check_properties(struct pw_impl_node *node)
 		}
 	}
 	node->lock_rate = pw_properties_get_bool(node->properties, PW_KEY_NODE_LOCK_RATE, false);
+	/* the leaf node is one that only produces/consumes the data. We can deduce this from the
+	 * absence of a link-group and the fact that it has no output/input ports. */
+	node->leaf = node->link_groups == NULL &&
+			(node->info.max_input_ports == 0 || node->info.max_output_ports == 0);
 
 	value = pw_properties_get_uint32(node->properties, PW_KEY_NODE_FORCE_RATE, SPA_ID_INVALID);
 	if (value == 0)
@@ -1782,7 +1788,6 @@ static int update_properties(struct pw_impl_node *node, const struct spa_dict *d
 	int changed;
 
 	changed = pw_properties_update_ignore(node->properties, dict, filter ? ignored : NULL);
-	node->info.props = &node->properties->dict;
 
 	pw_log_debug("%p: updated %d properties", node, changed);
 
@@ -2363,8 +2368,8 @@ void pw_impl_node_add_rt_listener(struct pw_impl_node *node,
 			   void *data)
 {
 	struct listener_data d = { .listener = listener, .events = events, .data = data };
-	pw_loop_invoke(node->data_loop,
-                       do_add_rt_listener, SPA_ID_INVALID, &d, sizeof(d), false, node);
+	pw_loop_locked(node->data_loop,
+                       do_add_rt_listener, SPA_ID_INVALID, &d, sizeof(d), node);
 }
 
 static int do_remove_listener(struct spa_loop *loop,
@@ -2379,8 +2384,8 @@ SPA_EXPORT
 void pw_impl_node_remove_rt_listener(struct pw_impl_node *node,
 			  struct spa_hook *listener)
 {
-	pw_loop_invoke(node->data_loop,
-                       do_remove_listener, SPA_ID_INVALID, NULL, 0, true, listener);
+	pw_loop_locked(node->data_loop,
+                       do_remove_listener, SPA_ID_INVALID, NULL, 0, listener);
 }
 
 /** Destroy a node
@@ -2711,6 +2716,19 @@ error:
 	return SPA_ID_INVALID;
 }
 
+SPA_EXPORT
+struct pw_impl_port *pw_impl_node_get_free_port(struct pw_impl_node *node, enum pw_direction direction)
+{
+	uint32_t port_id = pw_impl_node_get_free_port_id(node, direction);
+	if (port_id == SPA_ID_INVALID)
+		return NULL;
+
+	spa_node_add_port(node->node, direction, port_id, NULL);
+
+	return pw_impl_node_find_port(node, direction, port_id);
+}
+
+
 static void on_state_complete(void *obj, void *data, int res, uint32_t seq)
 {
 	struct pw_impl_node *node = obj;
