@@ -7,10 +7,16 @@
 
 #include <pipewire/properties.h>
 
+struct spa_hook;
+
 #ifdef __cplusplus
 extern "C" {
 #endif
 
+#ifndef PW_API_MEM
+#define PW_API_MEM static inline
+#endif
+
 /** \defgroup pw_memblock Memory Blocks
  * Memory allocation and pools.
  */
@@ -123,7 +129,7 @@ struct pw_memblock * pw_mempool_import(struct pw_mempool *pool,
 void pw_memblock_free(struct pw_memblock *mem);
 
 /** Unref a memblock */
-static inline void pw_memblock_unref(struct pw_memblock *mem)
+PW_API_MEM void pw_memblock_unref(struct pw_memblock *mem)
 {
 	if (--mem->ref == 0)
 		pw_memblock_free(mem);
@@ -173,7 +179,7 @@ struct pw_map_range {
 
 /** Calculate parameters to mmap() memory into \a range so that
  * \a size bytes at \a offset can be mapped with mmap().  */
-static inline void pw_map_range_init(struct pw_map_range *range,
+PW_API_MEM void pw_map_range_init(struct pw_map_range *range,
 				     uint32_t offset, uint32_t size,
 				     uint32_t page_size)
 {
