@@ -3,6 +3,8 @@
 /*                         @author Pantelis Antoniou <pantelis.antoniou@konsulko.com> */
 /* SPDX-License-Identifier: MIT */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <errno.h>
 #include <time.h>
@@ -34,15 +36,23 @@
 #include <pipewire/i18n.h>
 #include <pipewire/extensions/metadata.h>
 
-#include "config.h"
-
 #ifdef HAVE_PW_CAT_FFMPEG_INTEGRATION
 #include <libavformat/avformat.h>
 #include <libavcodec/avcodec.h>
 #include <libavutil/log.h>
+
+#ifndef AV_PROFILE_DTS_HD_MA
+#define AV_PROFILE_DTS_HD_MA FF_PROFILE_DTS_HD_MA
+#endif
+
+#ifndef AV_PROFILE_DTS_HD_HRA
+#define AV_PROFILE_DTS_HD_HRA FF_PROFILE_DTS_HD_HRA
+#endif
+
 #endif
 
 #include "midifile.h"
+#include "midiclip.h"
 #include "dfffile.h"
 #include "dsffile.h"
 
@@ -103,6 +113,8 @@ struct data {
 #ifdef HAVE_PW_CAT_FFMPEG_INTEGRATION
 #define TYPE_ENCODED    3
 #endif
+#define TYPE_SYSEX	4
+#define TYPE_MIDI2	5
 	int data_type;
 	bool raw;
 	const char *remote_name;
@@ -141,7 +153,15 @@ struct data {
 	struct {
 		struct midi_file *file;
 		struct midi_file_info info;
+#define MIDI_FORCE_NONE		0
+#define MIDI_FORCE_UMP		1
+#define MIDI_FORCE_MIDI1	2
+		int force_type;
 	} midi;
+	struct {
+		struct midi_clip *file;
+		struct midi_clip_info info;
+	} clip;
 	struct {
 		struct dsf_file *file;
 		struct dsf_file_info info;
@@ -162,6 +182,12 @@ struct data {
 		int64_t accumulated_excess_playtime;
 	} encoded;
 #endif
+	struct {
+		FILE *file;
+	} sysex;
+
+	uint64_t sample_limit;         /* 0 means unlimited */
+	uint64_t samples_processed;
 };
 
 #define STR_FMTS "(ulaw|alaw|u8|s8|s16|s32|f32|f64)"
@@ -249,8 +275,10 @@ static int encoded_playback_fill(struct data *d, void *dest, unsigned int n_fram
 {
 	AVPacket *packet = d->encoded.packet;
 	int ret;
+	uint8_t *dest_ptr = (uint8_t *)dest;
 	struct pw_time time;
 	int64_t quantum_duration;
+	int64_t accumulated_duration;
 	int64_t excess_playtime;
 	int64_t cycle_length;
 	int64_t av_time_base_num, av_time_base_denom;
@@ -298,32 +326,42 @@ static int encoded_playback_fill(struct data *d, void *dest, unsigned int n_fram
 		return 0;
 	}
 
-	/* Keep reading packets until we get one from the stream we are
-	 * interested in. This is relevant when playing data that contains
-	 * several multiplexed streams. */
-	while (true) {
-		if ((ret = av_read_frame(d->encoded.format_context, packet) < 0))
-			break;
+	accumulated_duration = 0;
 
-		if (packet->stream_index == d->encoded.stream_index)
+	/* Continue filling the buffer until at least a quantum's worth of data
+	 * has been written. Encoded frames may have a playtime that is _shorter_
+	 * than a quantum. In that case, multiple frames must be written into the
+	 * buffer to cover a quantum length. */
+	while (true) {
+		/* Keep reading packets until we get one from the stream we are
+		 * interested in. This is relevant when playing data that contains
+		 * several multiplexed streams. */
+		while (true) {
+			if ((ret = av_read_frame(d->encoded.format_context, packet) < 0))
+				break;
+
+			if (packet->stream_index == d->encoded.stream_index)
+				break;
+		}
+
+		memcpy(dest_ptr, packet->data, packet->size);
+
+		accumulated_duration += packet->duration;
+		dest_ptr += packet->size;
+
+		if (accumulated_duration >= quantum_duration)
+		{
+			excess_playtime = accumulated_duration - quantum_duration;
+			d->encoded.accumulated_excess_playtime += excess_playtime;
 			break;
+		}
 	}
 
-	memcpy(dest, packet->data, packet->size);
-
-	if (packet->duration > quantum_duration)
-		excess_playtime = packet->duration - quantum_duration;
-	else
-		excess_playtime = 0;
-	d->encoded.accumulated_excess_playtime += excess_playtime;
-
-	return packet->size;
+	return (dest_ptr - ((uint8_t*)dest));
 }
 
 static int av_codec_params_to_audio_info(struct data *data, AVCodecParameters *codec_params, struct spa_audio_info *info)
 {
-	int32_t profile;
-
 	switch (codec_params->codec_id) {
 	case AV_CODEC_ID_VORBIS:
 		info->media_subtype = SPA_MEDIA_SUBTYPE_vorbis;
@@ -348,32 +386,50 @@ static int av_codec_params_to_audio_info(struct data *data, AVCodecParameters *c
 	case AV_CODEC_ID_WMAVOICE:
 	case AV_CODEC_ID_WMALOSSLESS:
 		info->media_subtype = SPA_MEDIA_SUBTYPE_wma;
+		info->info.wma.rate = data->rate;
+		info->info.wma.channels = data->channels;
+		info->info.wma.bitrate = data->bitrate;
+		info->info.wma.block_align = codec_params->block_align;
+		/* The codec tag is not decided by FFmpeg; instead, it is
+		 * directly taken from the container. FFmpeg currently has
+		 * no named constants for the WMA versions, so numeric
+		 * constants have to be used here instead. */
 		switch (codec_params->codec_tag) {
-		/* TODO see if these hex constants can be replaced by named constants from FFmpeg */
+		/* This originates from Microsoft's mmreg.h , where
+		 * 0x160 is specified as meaning WMA v1 (which is commonly
+		 * referred to as WMA 7). */
+		case 0x160:
+			info->info.wma.profile = SPA_AUDIO_WMA_PROFILE_WMA7;
+			break;
+		/* This originates from Microsoft's mmreg.h , where
+		 * 0x161 is specified as meaning WMA v2 (which is commonly
+		 * referred to as WMA 8 or 9). */
 		case 0x161:
-			profile = SPA_AUDIO_WMA_PROFILE_WMA9;
+			info->info.wma.profile = SPA_AUDIO_WMA_PROFILE_WMA9;
 			break;
+		/* This originates from Microsoft's mmreg.h , where
+		 * 0x162 is specified as meaning WMA v3 (which is commonly
+		 * referred to as WMA 9 Pro). */
 		case 0x162:
-			profile = SPA_AUDIO_WMA_PROFILE_WMA9_PRO;
+			info->info.wma.profile = SPA_AUDIO_WMA_PROFILE_WMA9_PRO;
 			break;
+		/* This originates from Microsoft's mmreg.h , where
+		 * 0x163 is specified as meaning WMA 9 lossless. */
 		case 0x163:
-			profile = SPA_AUDIO_WMA_PROFILE_WMA9_LOSSLESS;
+			info->info.wma.profile = SPA_AUDIO_WMA_PROFILE_WMA9_LOSSLESS;
 			break;
+		/* WMA 10 is not mentioned in mmreg.h - these were empirically
+		 * determined instead. */
 		case 0x166:
-			profile = SPA_AUDIO_WMA_PROFILE_WMA10;
+			info->info.wma.profile = SPA_AUDIO_WMA_PROFILE_WMA10;
 			break;
 		case 0x167:
-			profile = SPA_AUDIO_WMA_PROFILE_WMA10_LOSSLESS;
+			info->info.wma.profile = SPA_AUDIO_WMA_PROFILE_WMA10_LOSSLESS;
 			break;
 		default:
 			fprintf(stderr, "error: invalid WMA profile\n");
 			return -EINVAL;
 		}
-		info->info.wma.rate = data->rate;
-		info->info.wma.channels = data->channels;
-		info->info.wma.bitrate = data->bitrate;
-		info->info.wma.block_align = codec_params->block_align;
-		info->info.wma.profile = profile;
 		break;
 	case AV_CODEC_ID_FLAC:
 		info->media_subtype = SPA_MEDIA_SUBTYPE_flac;
@@ -408,6 +464,41 @@ static int av_codec_params_to_audio_info(struct data *data, AVCodecParameters *c
 		info->info.amr.channels = data->channels;
 		info->info.amr.band_mode = SPA_AUDIO_AMR_BAND_MODE_WB;
 		break;
+	case AV_CODEC_ID_AC3:
+		info->media_subtype = SPA_MEDIA_SUBTYPE_ac3;
+		info->info.ac3.rate = data->rate;
+		info->info.ac3.channels = data->channels;
+		break;
+	case AV_CODEC_ID_EAC3:
+		info->media_subtype = SPA_MEDIA_SUBTYPE_eac3;
+		info->info.eac3.rate = data->rate;
+		info->info.eac3.channels = data->channels;
+		break;
+	case AV_CODEC_ID_TRUEHD:
+		info->media_subtype = SPA_MEDIA_SUBTYPE_truehd;
+		info->info.truehd.rate = data->rate;
+		info->info.truehd.channels = data->channels;
+		break;
+	case AV_CODEC_ID_DTS:
+		info->media_subtype = SPA_MEDIA_SUBTYPE_dts;
+		info->info.dts.rate = data->rate;
+		info->info.dts.channels = data->channels;
+		switch (codec_params->profile) {
+		case AV_PROFILE_DTS_HD_MA:
+			info->info.dts.ext_type = SPA_AUDIO_DTS_EXT_HD_MA;
+			break;
+		case AV_PROFILE_DTS_HD_HRA:
+			info->info.dts.ext_type = SPA_AUDIO_DTS_EXT_HD_HRA;
+			break;
+		default:
+			info->info.dts.ext_type = SPA_AUDIO_DTS_EXT_NONE;
+			break;
+		}
+		break;
+	case AV_CODEC_ID_MPEGH_3D_AUDIO:
+		info->media_subtype = SPA_MEDIA_SUBTYPE_mpegh;
+		info->info.mpegh.rate = data->rate;
+		break;
 	default:
 		fprintf(stderr, "Unsupported encoded media subtype\n");
 		return -EINVAL;
@@ -787,7 +878,7 @@ on_param_changed(void *userdata, uint32_t id, const struct spa_pod *param)
 	data->dff.layout.channels = info.info.dsd.channels;
 	data->dff.layout.lsb = info.info.dsd.bitorder == SPA_PARAM_BITORDER_lsb;
 
-	data->stride = data->dsf.layout.channels * SPA_ABS(data->dsf.layout.interleave);
+	data->stride = dsf_layout_stride(&data->dsf.layout);
 
 	if (data->verbose) {
 		fprintf(stderr, "DSD: channels:%d bitorder:%s interleave:%d stride:%d\n",
@@ -831,8 +922,20 @@ static void on_process(void *userdata)
 		 * fill callback actually returns number of bytes, not frames, since
 		 * this is encoded data. However, the calculations below still work
 		 * out because the stride is set to 1 in setup_encodedfile(). */
+		if (data->sample_limit > 0) {
+			uint64_t samples_left = data->sample_limit - data->samples_processed;
+			if (samples_left == 0) {
+				pw_main_loop_quit(data->loop);
+				return;
+			}
+			n_frames = SPA_MIN(n_frames, (int)samples_left);
+		}
+
 		n_fill_frames = data->fill(data, p, n_frames, &null_frame);
 
+		if (data->sample_limit > 0 && n_fill_frames > 0)
+			data->samples_processed += n_fill_frames;
+
 		if (null_frame) {
 			/* A null frame is not to be confused with the drain scenario.
 			 * In this case, we want to continue streaming, but in this
@@ -864,8 +967,20 @@ static void on_process(void *userdata)
 
 		n_frames = size / data->stride;
 
+		if (data->sample_limit > 0) {
+			uint64_t samples_left = data->sample_limit - data->samples_processed;
+			if (samples_left == 0) {
+				pw_main_loop_quit(data->loop);
+				return;
+			}
+			n_frames = SPA_MIN(n_frames, (int)samples_left);
+		}
+
 		n_fill_frames = data->fill(data, p, n_frames, &null_frame);
 
+		if (data->sample_limit > 0 && n_fill_frames > 0)
+			data->samples_processed += n_fill_frames;
+
 		have_data = true;
 	}
 
@@ -942,6 +1057,7 @@ static const struct option long_options[] = {
 	{ "midi",		no_argument,	   NULL, 'm' },
 	{ "dsd",		no_argument,	   NULL, 'd' },
 	{ "encoded",		no_argument,	   NULL, 'o' },
+	{ "sysex",		no_argument,	   NULL, 's' },
 
 	{ "remote",		required_argument, NULL, 'R' },
 
@@ -959,6 +1075,9 @@ static const struct option long_options[] = {
 	{ "volume",		required_argument, NULL, OPT_VOLUME },
 	{ "quality",		required_argument, NULL, 'q' },
 	{ "raw",		no_argument, NULL, 'a' },
+	{ "force-midi",		required_argument, NULL, 'M' },
+	{ "sample-count",	required_argument, NULL, 'n' },
+	{ "midi-clip",		no_argument, NULL, 'c' },
 
 	{ NULL, 0, NULL, 0 }
 };
@@ -1004,6 +1123,8 @@ static void show_usage(const char *name, bool is_error)
 	     "      --volume                          Stream volume 0-1.0 (default %.3f)\n"
 	     "  -q  --quality                         Resampler quality (0 - 15) (default %d)\n"
 	     "  -a, --raw                             RAW mode\n"
+	     "  -M, --force-midi                      Force midi format, one of \"midi\" or \"ump\", (default ump)\n"
+	     "  -n, --sample-count COUNT              Stop after COUNT samples\n"
 	     "\n"),
 	     DEFAULT_RATE,
 	     DEFAULT_CHANNELS,
@@ -1020,6 +1141,8 @@ static void show_usage(const char *name, bool is_error)
 #ifdef HAVE_PW_CAT_FFMPEG_INTEGRATION
 		     "  -o, --encoded                         Encoded mode\n"
 #endif
+		     "  -s, --sysex                           SysEx mode\n"
+		     "  -c, --midi-clip                       MIDI clip mode\n"
 		     "\n"), fp);
 	}
 }
@@ -1044,6 +1167,8 @@ static int midi_play(struct data *d, void *src, unsigned int n_frames, bool *nul
 	while (1) {
 		uint32_t frame;
 		struct midi_event ev;
+		uint64_t state = 0;
+		size_t size;
 
 		res = midi_file_next_time(d->midi.file, &ev.sec);
 		if (res <= 0) {
@@ -1063,32 +1188,48 @@ static int midi_play(struct data *d, void *src, unsigned int n_frames, bool *nul
 		midi_file_read_event(d->midi.file, &ev);
 
 		if (d->verbose)
-			midi_file_dump_event(stderr, &ev);
+			midi_event_dump(stderr, &ev);
+
+		size = ev.size;
 
 		if (ev.type == MIDI_EVENT_TYPE_MIDI1) {
-			size_t size;
-			uint8_t *data;
-			uint64_t state = 0;
 
-			if (ev.data[0] == 0xff)
+			if (size < 1 || ev.data[0] == 0xff)
 				continue;
 
-			data = ev.data;
-			size = ev.size;
+			if (d->midi.force_type == MIDI_FORCE_UMP) {
+				uint8_t *data = ev.data;
+				while (size > 0) {
+					uint32_t ump[4];
+					int ump_size = spa_ump_from_midi(&data, &size,
+							ump, sizeof(ump), 0, &state);
+					if (ump_size <= 0)
+						break;
 
-			while (size > 0) {
-				uint32_t ump[4];
-				int ump_size = spa_ump_from_midi(&data, &size,
-						ump, sizeof(ump), 0, &state);
-				if (ump_size <= 0)
-					break;
-
-				spa_pod_builder_control(&b, frame, SPA_CONTROL_UMP);
-				spa_pod_builder_bytes(&b, ump, ump_size);
+					spa_pod_builder_control(&b, frame, SPA_CONTROL_UMP);
+					spa_pod_builder_bytes(&b, ump, ump_size);
+				}
+			} else {
+				spa_pod_builder_control(&b, frame, SPA_CONTROL_Midi);
+				spa_pod_builder_bytes(&b, ev.data, ev.size);
 			}
 		} else if (ev.type == MIDI_EVENT_TYPE_UMP) {
-			spa_pod_builder_control(&b, frame, SPA_CONTROL_UMP);
-			spa_pod_builder_bytes(&b, ev.data, ev.size);
+			if (d->midi.force_type == MIDI_FORCE_MIDI1) {
+				const uint32_t *data = (const uint32_t*)ev.data;
+				while (size > 0) {
+					uint8_t ev[16];
+					int ev_size = spa_ump_to_midi(&data, &size,
+							ev, sizeof(ev), &state);
+					if (ev_size <= 0)
+						break;
+
+					spa_pod_builder_control(&b, frame, SPA_CONTROL_Midi);
+					spa_pod_builder_bytes(&b, ev, ev_size);
+				}
+			} else {
+				spa_pod_builder_control(&b, frame, SPA_CONTROL_UMP);
+				spa_pod_builder_bytes(&b, ev.data, ev.size);
+			}
 		}
 		else
 			continue;
@@ -1102,32 +1243,41 @@ static int midi_play(struct data *d, void *src, unsigned int n_frames, bool *nul
 
 static int midi_record(struct data *d, void *src, unsigned int n_frames, bool *null_frame)
 {
-	struct spa_pod *pod;
-	struct spa_pod_control *c;
-	uint32_t frame;
+	struct spa_pod_parser parser;
+	struct spa_pod_frame frame;
+	struct spa_pod_sequence seq;
+	const void *seq_body, *c_body;
+	struct spa_pod_control c;
+	uint32_t offset;
 
-	frame = d->clock_time;
+	offset = d->clock_time;
 	d->clock_time += d->position->clock.duration;
 
-	if ((pod = spa_pod_from_data(src, n_frames, 0, n_frames)) == NULL)
-		return 0;
-	if (!spa_pod_is_sequence(pod))
+	spa_pod_parser_init_from_data(&parser, src, n_frames, 0, n_frames);
+
+	if (spa_pod_parser_push_sequence_body(&parser, &frame, &seq, &seq_body) < 0)
 		return 0;
 
-	SPA_POD_SEQUENCE_FOREACH((struct spa_pod_sequence*)pod, c) {
+	while (spa_pod_parser_get_control_body(&parser, &c, &c_body) >= 0) {
 		struct midi_event ev;
 
-		if (c->type != SPA_CONTROL_UMP)
+		switch (c.type) {
+		case SPA_CONTROL_UMP:
+			ev.type = MIDI_EVENT_TYPE_UMP;
+			break;
+		case SPA_CONTROL_Midi:
+			ev.type = MIDI_EVENT_TYPE_MIDI1;
+			break;
+		default:
 			continue;
-
+		}
 		ev.track = 0;
-		ev.sec = (frame + c->offset) / (float) d->position->clock.rate.denom;
-		ev.data = SPA_POD_BODY(&c->value),
-		ev.size = SPA_POD_BODY_SIZE(&c->value);
-		ev.type = MIDI_EVENT_TYPE_UMP;
+		ev.sec = (offset + c.offset) / (float) d->position->clock.rate.denom;
+		ev.data = (uint8_t*)c_body;
+		ev.size = c.value.size;
 
 		if (d->verbose)
-			midi_file_dump_event(stderr, &ev);
+			midi_event_dump(stderr, &ev);
 
 		midi_file_write_event(d->midi.file, &ev);
 	}
@@ -1163,6 +1313,184 @@ static int setup_midifile(struct data *data)
 	return 0;
 }
 
+static int clip_play(struct data *d, void *src, unsigned int n_frames, bool *null_frame)
+{
+	int res;
+	struct spa_pod_builder b;
+	struct spa_pod_frame f;
+	uint32_t first_frame, last_frame;
+	bool have_data = false;
+
+	spa_zero(b);
+	spa_pod_builder_init(&b, src, n_frames);
+
+	spa_pod_builder_push_sequence(&b, &f, 0);
+
+	first_frame = d->clock_time;
+	last_frame = first_frame + d->position->clock.duration;
+	d->clock_time = last_frame;
+
+	while (1) {
+		uint32_t frame;
+		struct midi_event ev;
+		uint64_t state = 0;
+		size_t size;
+
+		res = midi_clip_next_time(d->clip.file, &ev.sec);
+		if (res <= 0) {
+			if (have_data)
+				break;
+			return res;
+		}
+
+		frame = (uint32_t)(ev.sec * d->position->clock.rate.denom);
+		if (frame < first_frame)
+			frame = 0;
+		else if (frame < last_frame)
+			frame -= first_frame;
+		else
+			break;
+
+		midi_clip_read_event(d->clip.file, &ev);
+
+		if (d->verbose)
+			midi_event_dump(stderr, &ev);
+
+		size = ev.size;
+
+		if (d->midi.force_type == MIDI_FORCE_MIDI1) {
+			const uint32_t *data = (const uint32_t*)ev.data;
+			while (size > 0) {
+				uint8_t ev[16];
+				int ev_size = spa_ump_to_midi(&data, &size,
+						ev, sizeof(ev), &state);
+				if (ev_size <= 0)
+					break;
+
+				spa_pod_builder_control(&b, frame, SPA_CONTROL_Midi);
+				spa_pod_builder_bytes(&b, ev, ev_size);
+			}
+		} else {
+			spa_pod_builder_control(&b, frame, SPA_CONTROL_UMP);
+			spa_pod_builder_bytes(&b, ev.data, ev.size);
+		}
+		have_data = true;
+	}
+	spa_pod_builder_pop(&b, &f);
+
+	return b.state.offset;
+}
+
+static int clip_record(struct data *d, void *src, unsigned int n_frames, bool *null_frame)
+{
+	struct spa_pod_parser parser;
+	struct spa_pod_frame frame;
+	struct spa_pod_sequence seq;
+	const void *seq_body, *c_body;
+	struct spa_pod_control c;
+	uint32_t offset;
+
+	offset = d->clock_time;
+	d->clock_time += d->position->clock.duration;
+
+	spa_pod_parser_init_from_data(&parser, src, n_frames, 0, n_frames);
+
+	if (spa_pod_parser_push_sequence_body(&parser, &frame, &seq, &seq_body) < 0)
+		return 0;
+
+	while (spa_pod_parser_get_control_body(&parser, &c, &c_body) >= 0) {
+		struct midi_event ev;
+
+		switch (c.type) {
+		case SPA_CONTROL_UMP:
+			ev.type = MIDI_EVENT_TYPE_UMP;
+			break;
+		case SPA_CONTROL_Midi:
+			ev.type = MIDI_EVENT_TYPE_MIDI1;
+			break;
+		default:
+			continue;
+		}
+		ev.track = 0;
+		ev.sec = (offset + c.offset) / (float) d->position->clock.rate.denom;
+		ev.data = (uint8_t*)c_body;
+		ev.size = c.value.size;
+
+		if (d->verbose)
+			midi_event_dump(stderr, &ev);
+
+		midi_clip_write_event(d->clip.file, &ev);
+	}
+	return 0;
+}
+
+static int setup_midiclip(struct data *data)
+{
+	if (data->mode == mode_record) {
+		spa_zero(data->clip.info);
+		data->clip.info.format = 0;
+	}
+
+	data->clip.file = midi_clip_open(data->filename,
+			data->mode == mode_playback ? "r" : "w",
+			&data->clip.info);
+	if (data->clip.file == NULL) {
+		fprintf(stderr, "midiclip: can't read midi file '%s': %m\n", data->filename);
+		return -errno;
+	}
+
+	if (data->verbose)
+		fprintf(stderr, "midifile: opened file \"%s\" format %08x div:%d\n",
+				data->filename,
+				data->clip.info.format, data->clip.info.division);
+
+	data->fill = data->mode == mode_playback ?  clip_play : clip_record;
+	data->stride = 1;
+
+	return 0;
+}
+
+static int sysex_play(struct data *d, void *dst, unsigned int n_frames, bool *null_frame)
+{
+	struct spa_pod_builder b;
+	struct spa_pod_frame f;
+	size_t size, to_read = n_frames - 64;
+	uint8_t bytes[to_read];
+
+	spa_zero(b);
+	spa_pod_builder_init(&b, dst, n_frames);
+
+	spa_pod_builder_push_sequence(&b, &f, 0);
+	spa_pod_builder_control(&b, 0, SPA_CONTROL_Midi);
+
+	size = fread(bytes, 1, to_read, d->sysex.file);
+
+	spa_pod_builder_bytes(&b, bytes, size);
+	spa_pod_builder_pop(&b, &f);
+
+	return b.state.offset;
+}
+
+static int setup_sysex(struct data *data)
+{
+	if (data->mode == mode_record)
+		return -ENOTSUP;
+
+	data->sysex.file = fopen(data->filename, "r");
+	if (data->sysex.file == NULL) {
+		fprintf(stderr, "sysex: can't read file '%s': %m\n", data->filename);
+		return -errno;
+	}
+
+	if (data->verbose)
+		fprintf(stderr, "sysex: opened file \"%s\"\n", data->filename);
+
+	data->fill = sysex_play;
+	data->stride = 1;
+
+	return 0;
+}
+
 struct dsd_layout_info {
 	 uint32_t type;
 	 struct spa_audio_layout_info info;
@@ -1677,6 +2005,15 @@ int main(int argc, char *argv[])
 	} else if (spa_streq(prog, "pw-midirecord")) {
 		data.mode = mode_record;
 		data.data_type = TYPE_MIDI;
+	} else if (spa_streq(prog, "pw-midi2play")) {
+		data.mode = mode_playback;
+		data.data_type = TYPE_MIDI2;
+	} else if (spa_streq(prog, "pw-midi2record")) {
+		data.mode = mode_record;
+		data.data_type = TYPE_MIDI2;
+	} else if (spa_streq(prog, "pw-sysex")) {
+		data.mode = mode_playback;
+		data.data_type = TYPE_SYSEX;
 	} else if (spa_streq(prog, "pw-dsdplay")) {
 		data.mode = mode_playback;
 		data.data_type = TYPE_DSD;
@@ -1691,6 +2028,7 @@ int main(int argc, char *argv[])
 	/* negative means no volume adjustment */
 	data.volume = -1.0;
 	data.quality = -1;
+	data.midi.force_type = MIDI_FORCE_UMP;
 	data.props = pw_properties_new(
 			PW_KEY_APP_NAME, prog,
 			PW_KEY_NODE_NAME, prog,
@@ -1702,9 +2040,9 @@ int main(int argc, char *argv[])
 	}
 
 #ifdef HAVE_PW_CAT_FFMPEG_INTEGRATION
-	while ((c = getopt_long(argc, argv, "hvprmdoR:q:P:a", long_options, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, "hvprmdosR:q:P:aM:n:c", long_options, NULL)) != -1) {
 #else
-	while ((c = getopt_long(argc, argv, "hvprmdR:q:P:a", long_options, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, "hvprmdsR:q:P:aM:n:c", long_options, NULL)) != -1) {
 #endif
 
 		switch (c) {
@@ -1747,6 +2085,9 @@ int main(int argc, char *argv[])
 			data.data_type = TYPE_ENCODED;
 			break;
 #endif
+		case 's':
+			data.data_type = TYPE_SYSEX;
+			break;
 
 		case 'R':
 			data.remote_name = optarg;
@@ -1760,6 +2101,17 @@ int main(int argc, char *argv[])
 			data.raw = true;
 			break;
 
+		case 'M':
+			if (spa_streq(optarg, "midi"))
+				data.midi.force_type = MIDI_FORCE_MIDI1;
+			else if (spa_streq(optarg, "ump"))
+				data.midi.force_type = MIDI_FORCE_UMP;
+			else {
+				fprintf(stderr, "error: bad force-midi %s\n", optarg);
+				goto error_usage;
+			}
+			break;
+
 		case OPT_MEDIA_TYPE:
 			data.media_type = optarg;
 			break;
@@ -1823,6 +2175,12 @@ int main(int argc, char *argv[])
 			if (!spa_atof(optarg, &data.volume))
 				data.volume = (float)atof(optarg);
 			break;
+		case 'n':
+			data.sample_limit = strtoull(optarg, NULL, 10);
+			break;
+		case 'c':
+			data.data_type = TYPE_MIDI2;
+			break;
 		default:
 			goto error_usage;
 		}
@@ -1836,6 +2194,8 @@ int main(int argc, char *argv[])
 	if (!data.media_type) {
 		switch (data.data_type) {
 		case TYPE_MIDI:
+		case TYPE_MIDI2:
+		case TYPE_SYSEX:
 			data.media_type = DEFAULT_MIDI_MEDIA_TYPE;
 			break;
 		default:
@@ -1915,6 +2275,9 @@ int main(int argc, char *argv[])
 		case TYPE_MIDI:
 			ret = setup_midifile(&data);
 			break;
+		case TYPE_MIDI2:
+			ret = setup_midiclip(&data);
+			break;
 		case TYPE_DSD:
 			ret = setup_dsdfile(&data);
 			break;
@@ -1923,6 +2286,9 @@ int main(int argc, char *argv[])
 			ret = setup_encodedfile(&data);
 			break;
 #endif
+		case TYPE_SYSEX:
+			ret = setup_sysex(&data);
+			break;
 		default:
 			ret = -ENOTSUP;
 			break;
@@ -1985,6 +2351,8 @@ int main(int argc, char *argv[])
 		break;
 	}
 	case TYPE_MIDI:
+	case TYPE_MIDI2:
+	case TYPE_SYSEX:
 		params[n_params++] = spa_pod_builder_add_object(&b,
 				SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat,
 				SPA_FORMAT_mediaType,		SPA_POD_Id(SPA_MEDIA_TYPE_application),
@@ -2106,6 +2474,8 @@ error_no_main_loop:
 		sf_close(data.file);
 	if (data.midi.file)
 		midi_file_close(data.midi.file);
+	if (data.clip.file)
+		midi_clip_close(data.clip.file);
 	if (data.dsf.file)
 		dsf_file_close(data.dsf.file);
 	if (data.dff.file)
